# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import io
import struct
import asyncio as aio
from .tlv_type import BinaryStr, VarBinaryStr


__all__ = ['get_tl_num_size', 'write_tl_num', 'pack_uint_bytes', 'parse_tl_num', 'read_tl_num_from_stream',
           'parse_and_check_tl']


def get_tl_num_size(val: int) -> int:
    """
    Calculate the length of a TL variable.

    :param val: an integer standing for Type or Length.
    :return: The length of var.
    """
    if val <= 0xFC:
        return 1
    elif val <= 0xFFFF:
        return 3
    elif val <= 0xFFFFFFFF:
        return 5
    else:
        return 9


def write_tl_num(val: int, buf: VarBinaryStr, offset: int = 0) -> int:
    """
    Write a Type or Length value into a buffer.

    :param val: the value.
    :param buf: the buffer.
    :param offset: the starting offset.
    :return: the encoded length.
    """
    if val <= 0xFC:
        struct.pack_into('!B', buf, offset, val)
        return 1
    elif val <= 0xFFFF:
        struct.pack_into('!BH', buf, offset, 0xFD, val)
        return 3
    elif val <= 0xFFFFFFFF:
        struct.pack_into('!BI', buf, offset, 0xFE, val)
        return 5
    else:
        struct.pack_into('!BQ', buf, offset, 0xFF, val)
        return 9


def pack_uint_bytes(val: int) -> bytes:
    """
    Pack an non-negative integer value into bytes

    :param val: the integer.
    :return: the buffer.
    """
    if val <= 0xFF:
        return struct.pack('!B', val)
    elif val <= 0xFFFF:
        return struct.pack('!H', val)
    elif val <= 0xFFFFFFFF:
        return struct.pack('!I', val)
    else:
        return struct.pack('!Q', val)


def parse_tl_num(buf: BinaryStr, offset: int = 0) -> (int, int):
    """
    Parse a Type or Length variable from a buffer.

    :param buf: the buffer.
    :param offset: the starting offset.
    :return: a pair (value, size parsed).
    """
    ret = buf[offset]
    if ret <= 0xFC:
        return ret, 1
    elif ret == 0xFD:
        return struct.unpack('!H', buf[offset+1:offset+3])[0], 3
    elif ret == 0xFE:
        return struct.unpack('!I', buf[offset+1:offset+5])[0], 5
    else:
        return struct.unpack('!Q', buf[offset+1:offset+9])[0], 9


async def read_tl_num_from_stream(reader: aio.StreamReader, bio: io.BytesIO) -> int:
    """
    Read a Type or Length variable from a StreamReader.

    :param reader: the StreamReader.
    :param bio: the BytesIO to write whatever is read from the stream.
    :return: the value read.
    """
    buf = await reader.readexactly(1)
    bio.write(buf)
    num = buf[0]
    if num <= 0xFC:
        return num
    elif num == 0xFD:
        buf = await reader.readexactly(2)
        bio.write(buf)
        return struct.unpack('!H', buf)[0]
    elif num == 0xFE:
        buf = await reader.readexactly(4)
        bio.write(buf)
        return struct.unpack('!I', buf)[0]
    else:
        buf = await reader.readexactly(8)
        bio.write(buf)
        return struct.unpack('!Q', buf)[0]


def parse_and_check_tl(wire: BinaryStr, expected_type: int) -> memoryview:
    """
    Parse Type and Length, and then check:

    - If the Type equals `expected_type`;
    - If the Length equals the length of `wire`.

    :param wire: the TLV encoded wire.
    :param expected_type: expected Type.
    :return: a pointer to the memory of Value.
    """
    typ, typ_len = parse_tl_num(wire, 0)
    size, siz_len = parse_tl_num(wire, typ_len)
    if typ != expected_type:
        raise ValueError(f'wire is of type {typ} but {expected_type} is expected')
    if len(wire) != typ_len+siz_len+size:
        raise IndexError(f'wire size {len(wire)} mismatch with object size {size}')
    return memoryview(wire)[typ_len+siz_len:typ_len+siz_len+size]


def shrink_length(wire: VarBinaryStr, val: int) -> VarBinaryStr:
    # assert val > 0
    wire = memoryview(wire)
    typ, typ_len = parse_tl_num(wire, 0)
    size, siz_len = parse_tl_num(wire, typ_len)
    real_size = size - val
    new_siz_len = write_tl_num(real_size, wire, typ_len)
    if new_siz_len == siz_len:
        return wire[:-val]
    else:
        diff = siz_len - new_siz_len
        write_tl_num(typ, wire, diff)
        write_tl_num(real_size, wire, typ_len + diff)
        return wire[diff:-val]



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Union, List, Iterable


__all__ = ['BinaryStr', 'VarBinaryStr', 'FormalName', 'NonStrictName', 'is_binary_str']


BinaryStr = Union[bytes, bytearray, memoryview]
r"""A binary string is any of :class:`bytes`, :class:`bytearray`, :class:`memoryview`."""

VarBinaryStr = Union[bytearray, memoryview]
r"""A variant binary string is a :class:`bytearray` or a non-readonly :class:`memoryview`."""

FormalName = List[BinaryStr]
r"""A FormalName is a list of encoded Components."""

NonStrictName = Union[Iterable[Union[BinaryStr, str]], str, BinaryStr]
r"""
A NonStrictName is any of below:

- A URI string.
- A list or iterator of Components, in the form of either encoded TLV or URI string.
- An encoded Name of type :class:`bytes`, :class:`bytearray` or :class:`memoryview`.

See also :ref:`label-different-names`
"""


def is_binary_str(var):
    r"""
    Check whether var is of type BinaryStr.

    :param var: The variable to check.
    :return: ``True`` if var is a :any:`BinaryStr`.
    """
    return isinstance(var, bytes) or isinstance(var, bytearray) or isinstance(var, memoryview)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from .keychain import Keychain
from typing import Dict, Any
from ..signer.sha256_digest_signer import DigestSha256Signer


class KeychainDigest(Keychain):
    """
    A signer which has no Identity and always returns a SHA-256 digest signer.
    """
    def get_signer(self, sign_args: Dict[str, Any]):
        if sign_args.pop('no_signature', False):
            return None
        else:
            return DigestSha256Signer()

    def __iter__(self):
        return None

    def __len__(self) -> int:
        return 0

    def __getitem__(self, name):
        raise KeyError(name)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import asyncio as aio
from ndn.encoding import parse_tl_num
from ndn.transport.stream_socket import Face


class DummyFace(Face):
    app = None

    def __init__(self, test_func):
        super().__init__()
        self.output_buf = b''
        self.test_func = test_func
        self.event = aio.Event()
        self.expected_len = 2 ** 32

    async def open(self):
        self.running = True

    def shutdown(self):
        self.running = False

    def send(self, data: bytes):
        self.output_buf += data
        if len(self.output_buf) >= self.expected_len:
            self.event.set()

    async def run(self):
        await self.test_func(self)
        if self.app:
            self.app.shutdown()

    async def consume_output(self, expected_output, timeout=0.01):
        self.expected_len = len(expected_output)
        if len(self.output_buf) < self.expected_len:
            await aio.wait_for(self.event.wait(), timeout)
        self.expected_len = 2 ** 32
        self.event.clear()
        assert self.output_buf == expected_output
        self.output_buf = b''

    async def ignore_output(self, length, timeout=0.1):
        self.expected_len = length
        await aio.wait_for(self.event.wait(), timeout)
        self.expected_len = 2 ** 32
        self.event.clear()
        self.output_buf = b''

    async def input_packet(self, packet):
        packet = memoryview(packet)
        typ, typ_len = parse_tl_num(packet)
        siz, siz_len = parse_tl_num(packet, typ_len)
        offset = typ_len + siz_len
        assert len(packet) == offset + siz
        await self.callback(typ, packet)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import os
import sys
from configparser import ConfigParser
from .security import TpmFile, Keychain, KeychainSqlite3
from .transport.stream_socket import Face, UnixFace, TcpFace
if sys.platform == 'darwin':
    from .security.tpm.tpm_osx_keychain import TpmOsxKeychain


def read_client_conf():
    def get_path():
        path = os.path.expanduser('~/.ndn/client.conf')
        if os.path.exists(path):
            return path
        path = '/usr/local/etc/ndn/client.conf'
        if os.path.exists(path):
            return path
        path = '/opt/local/etc/ndn/client.conf'
        if os.path.exists(path):
            return path
        path = '/etc/ndn/client.conf'
        if os.path.exists(path):
            return path

    def resolve_location(value):
        nonlocal path
        sp = value.split(':')
        if len(sp) == 1:
            schema = value
            loc = ''
        else:
            schema, loc = sp
        if not loc or not os.path.exists(loc):
            if loc and (path is not None):
                loc = os.path.join(os.path.dirname(path), loc)
            if not loc or not os.path.exists(loc):
                loc = '~/.ndn/ndnsec-key-file' if schema == 'tpm-file' else '~/.ndn'
                loc = os.path.expanduser(loc)
        return ':'.join((schema, loc))

    path = get_path()
    ret = {
        'transport': 'unix:///run/nfd.sock' if sys.platform == 'linux' else 'unix:///var/run/nfd.sock',
        'pib': 'pib-sqlite3',
        'tpm': 'tpm-osxkeychain' if sys.platform == 'darwin' else 'tpm-file'
    }
    if path:
        parser = ConfigParser()
        text = '[DEFAULT]\n'
        with open(path) as f:
            text += f.read()
        parser.read_string(text)
        for key in ret.keys():
            try:
                ret[key] = parser['DEFAULT'][key]
            except KeyError:
                pass
    for key in ret.keys():
        try:
            ret[key] = os.environ[f'NDN_CLIENT_{key.upper()}']
        except KeyError:
            pass
    for key in ['pib', 'tpm']:
        ret[key] = resolve_location(ret[key])
    return ret


def default_keychain(pib: str, tpm: str) -> Keychain:
    pib_schema, pib_loc = pib.split(':')
    tpm_schema, tpm_loc = tpm.split(':')
    if tpm_schema == 'tpm-file':
        tpm = TpmFile(tpm_loc)
    elif tpm_schema == 'tpm-osxkeychain':
        tpm = TpmOsxKeychain()
    else:
        raise ValueError(f'Unrecognized tpm schema: {tpm}')
    if pib_schema == 'pib-sqlite3':
        pib = KeychainSqlite3(os.path.join(pib_loc, 'pib.db'), tpm)
    else:
        raise ValueError(f'Unrecognized pib schema: {pib}')
    return pib


def default_face(face: str) -> Face:
    schema, uri = face.split('://')
    if schema == 'unix':
        return UnixFace(uri)
    elif schema == 'tcp' or schema == 'tcp4':
        if uri.find(':') >= 0:
            host, port = uri.split(':')
            port = port
        else:
            host = uri
            port = 6363
        return TcpFace(host, int(port))
    else:
        raise ValueError(f'Unrecognized face: {face}')



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import dataclasses as dc
from hashlib import sha256
from typing import Optional, List, Tuple
from .name import Name, Component
from .signer import Signer
from .tlv_type import VarBinaryStr, BinaryStr, NonStrictName, FormalName
from .tlv_var import parse_and_check_tl, shrink_length
from .tlv_model import TlvModel, InterestNameField, BoolField, UintField, \
    SignatureValueField, OffsetMarker, BytesField, ModelField, NameField, \
    ProcedureArgument, RepeatedField


__all__ = ['TypeNumber', 'ContentType', 'SignatureType', 'KeyLocator', 'SignatureInfo', 'Delegation',
           'Links', 'MetaInfo', 'InterestParam', 'SignaturePtrs', 'make_interest', 'make_data',
           'parse_interest', 'parse_data', 'Interest', 'Data']


class TypeNumber:
    r"""
    TLV Type numbers used in `NDN Packet Format 0.3
    <https://named-data.net/doc/NDN-packet-spec/current/types.html>`_.

    Constant names are changed to PEP 8 style, i.e., all upper cases with underscores separating words.
    """
    INTEREST = 0x05
    DATA = 0x06

    NAME = Name.TYPE_NAME
    GENERIC_NAME_COMPONENT = Component.TYPE_GENERIC
    IMPLICIT_SHA256_DIGEST_COMPONENT = Component.TYPE_IMPLICIT_SHA256
    PARAMETERS_SHA256_DIGEST_COMPONENT = Component.TYPE_PARAMETERS_SHA256

    CAN_BE_PREFIX = 0x21
    MUST_BE_FRESH = 0x12
    FORWARDING_HINT = 0x1e
    NONCE = 0x0a
    INTEREST_LIFETIME = 0x0c
    HOP_LIMIT = 0x22
    APPLICATION_PARAMETERS = 0x24
    INTEREST_SIGNATURE_INFO = 0x2c
    INTEREST_SIGNATURE_VALUE = 0x2e

    META_INFO = 0x14
    CONTENT = 0x15
    SIGNATURE_INFO = 0x16
    SIGNATURE_VALUE = 0x17
    CONTENT_TYPE = 0x18
    FRESHNESS_PERIOD = 0x19
    FINAL_BLOCK_ID = 0x1a

    SIGNATURE_TYPE = 0x1b
    KEY_LOCATOR = 0x1c
    KEY_DIGEST = 0x1d
    SIGNATURE_NONCE = 0x26
    SIGNATURE_TIME = 0x28
    SIGNATURE_SEQ_NUM = 0x2a

    DELEGATION = 0x1f
    PREFERENCE = 0x1e


class ContentType:
    r"""
    Numbers used in ContentType.

    ====   ===================================
    Type   Description
    ====   ===================================
    BLOB   Payload identified by the data name
    LINK   A list of delegations
    KEY    Public Key
    NACK   Application-level NACK
    ====   ===================================
    """
    BLOB = 0
    LINK = 1
    KEY = 2
    NACK = 3


class SignatureType:
    r"""
    Numbers used in SignatureType.

    =================   ==============================================
    Type                Description
    =================   ==============================================
    NOT_SIGNED          Not signed
    DIGEST_SHA256       SHA-256 digest (only for integrity protection)
    SHA256_WITH_RSA     RSA signature over a SHA-256 digest
    SHA256_WITH_ECDSA   An ECDSA signature over a SHA-256 digest
    HMAC_WITH_SHA256    SHA256 hash-based message authentication codes
    =================   ==============================================
    """
    NOT_SIGNED = None
    DIGEST_SHA256 = 0
    SHA256_WITH_RSA = 1
    SHA256_WITH_ECDSA = 3
    HMAC_WITH_SHA256 = 4


class KeyLocator(TlvModel):
    name = NameField()
    key_digest = BytesField(TypeNumber.KEY_DIGEST)


class SignatureInfo(TlvModel):
    signature_type = UintField(TypeNumber.SIGNATURE_TYPE, fixed_len=1)
    key_locator = ModelField(TypeNumber.KEY_LOCATOR, KeyLocator)
    signature_nonce = UintField(TypeNumber.SIGNATURE_NONCE)
    signature_time = UintField(TypeNumber.SIGNATURE_TIME)
    signature_seq_num = UintField(TypeNumber.SIGNATURE_SEQ_NUM)


class Delegation(TlvModel):
    preference = UintField(TypeNumber.PREFERENCE)
    delegation = NameField()


class Links(TlvModel):
    delegations = RepeatedField(ModelField(TypeNumber.DELEGATION, Delegation))


class InterestPacketValue(TlvModel):
    _signer = ProcedureArgument()
    _sig_cover_part = ProcedureArgument()
    _sig_value_buf = ProcedureArgument()
    _need_digest = ProcedureArgument()
    _digest_cover_part = ProcedureArgument()
    _digest_buf = ProcedureArgument()
    _shrink_len = ProcedureArgument(0)

    name = InterestNameField(need_digest=_need_digest,
                             signature_covered_part=_sig_cover_part,
                             digest_buffer=_digest_buf,
                             default="/")
    can_be_prefix = BoolField(TypeNumber.CAN_BE_PREFIX, default=False)
    must_be_fresh = BoolField(TypeNumber.MUST_BE_FRESH, default=False)
    forwarding_hint = ModelField(TypeNumber.FORWARDING_HINT, Links)
    nonce = UintField(TypeNumber.NONCE, fixed_len=4)
    lifetime = UintField(TypeNumber.INTEREST_LIFETIME)  # We can not write 4000 as a parse default
    hop_limit = UintField(TypeNumber.HOP_LIMIT, fixed_len=1)
    _sig_cover_start = OffsetMarker()
    _digest_cover_start = OffsetMarker()
    application_parameters = BytesField(TypeNumber.APPLICATION_PARAMETERS)
    signature_info = ModelField(TypeNumber.INTEREST_SIGNATURE_INFO, SignatureInfo)
    signature_value = SignatureValueField(TypeNumber.INTEREST_SIGNATURE_VALUE,
                                          signer=_signer,
                                          covered_part=_sig_cover_part,
                                          starting_point=_sig_cover_start,
                                          value_buffer=_sig_value_buf,
                                          shrink_len=_shrink_len)
    _digest_cover_end = OffsetMarker()

    def encoded_length(self, markers: Optional[dict] = None) -> int:
        if markers is None:
            markers = {}
        self._sig_cover_part.set_arg(markers, [])

        signer = self._signer.get_arg(markers)
        if signer is not None:
            signer.write_signature_info(self.signature_info)
        app_param = self.application_parameters
        if (signer is not None) and (app_param is None):
            app_param = b''
            self.application_parameters = app_param

        self._need_digest.set_arg(markers, app_param is not None)

        return super().encoded_length(markers)

    def encode(self,
               wire: VarBinaryStr = None,
               offset: int = 0,
               markers: Optional[dict] = None) -> VarBinaryStr:
        if markers is None:
            markers = {}
        ret = super().encode(wire, offset, markers)
        wire_view = memoryview(ret)

        InterestPacketValue.signature_value.calculate_signature(markers)
        if self._need_digest.get_arg(markers):
            digest_cover_start = self._digest_cover_start.get_arg(markers)
            shrink_size = self._shrink_len.get_arg(markers)
            digest_cover_end = self._digest_cover_end.get_arg(markers) - shrink_size
            digest_covered_part = [wire_view[digest_cover_start:digest_cover_end]]
            self._digest_cover_part.set_arg(markers, digest_covered_part)
            sha256_algo = sha256()
            digest_buf = self._digest_buf.get_arg(markers)
            for blk in digest_covered_part:
                sha256_algo.update(blk)
            digest_buf[:] = sha256_algo.digest()

        return ret

    @classmethod
    def parse(cls, wire: BinaryStr, markers: Optional[dict] = None, ignore_critical: bool = False):
        if markers is None:
            markers = {}
        cls._sig_cover_part.set_arg(markers, [])
        ret = super().parse(wire, markers, ignore_critical)
        digest_cover_start = cls._digest_cover_start.get_arg(markers)
        digest_cover_end = cls._digest_cover_end.get_arg(markers)
        digest_cover_part = [memoryview(wire)[digest_cover_start:digest_cover_end]]
        cls._digest_cover_part.set_arg(markers, digest_cover_part)
        return ret


class InterestPacket(TlvModel):
    _signer = ProcedureArgument()
    interest = ModelField(TypeNumber.INTEREST, InterestPacketValue, [_signer])


class MetaInfo(TlvModel):
    content_type = UintField(TypeNumber.CONTENT_TYPE)
    freshness_period = UintField(TypeNumber.FRESHNESS_PERIOD)
    final_block_id = BytesField(TypeNumber.FINAL_BLOCK_ID)

    def __init__(self,
                 content_type: int = ContentType.BLOB,
                 freshness_period: Optional[int] = None,
                 final_block_id: BinaryStr = None):
        self.content_type = content_type
        self.freshness_period = freshness_period
        self.final_block_id = final_block_id

    @staticmethod
    def from_dict(kwargs):
        return MetaInfo(**{f.name: kwargs[f.name]
                           for f in MetaInfo._encoded_fields
                           if f.name in kwargs})


class DataPacketValue(TlvModel):
    _signer = ProcedureArgument()
    _sig_cover_part = ProcedureArgument()
    _sig_value_buf = ProcedureArgument()
    _shrink_len = ProcedureArgument(0)

    _sig_cover_start = OffsetMarker()
    name = NameField("/")
    meta_info = ModelField(TypeNumber.META_INFO, MetaInfo)
    content = BytesField(TypeNumber.CONTENT)
    # v0.2 Data packets has critical SignatureType-specific TLVs
    signature_info = ModelField(TypeNumber.SIGNATURE_INFO, SignatureInfo, ignore_critical=True)
    signature_value = SignatureValueField(TypeNumber.SIGNATURE_VALUE,
                                          signer=_signer,
                                          covered_part=_sig_cover_part,
                                          starting_point=_sig_cover_start,
                                          value_buffer=_sig_value_buf,
                                          shrink_len=_shrink_len)

    def encoded_length(self, markers: Optional[dict] = None) -> int:
        if markers is None:
            markers = {}
        self._sig_cover_part.set_arg(markers, [])

        signer = self._signer.get_arg(markers)
        if signer is not None:
            signer.write_signature_info(self.signature_info)

        return super().encoded_length(markers)

    def encode(self,
               wire: VarBinaryStr = None,
               offset: int = 0,
               markers: Optional[dict] = None) -> VarBinaryStr:
        if markers is None:
            markers = {}
        ret = super().encode(wire, offset, markers)
        DataPacketValue.signature_value.calculate_signature(markers)
        return ret

    @classmethod
    def parse(cls, wire: BinaryStr, markers: Optional[dict] = None, ignore_critical: bool = False):
        if markers is None:
            markers = {}
        cls._sig_cover_part.set_arg(markers, [])
        return super().parse(wire, markers, ignore_critical)


class DataPacket(TlvModel):
    _signer = ProcedureArgument()
    data = ModelField(TypeNumber.DATA, DataPacketValue, [_signer])


@dc.dataclass
class InterestParam:
    r"""
    A dataclass collecting the parameters of an Interest, except ApplicationParameters.

    :ivar can_be_prefix: CanBePrefix. ``False`` by default.
    :vartype can_be_prefix: bool

    :ivar must_be_fresh: MustBeFresh. ``False`` by default.
    :vartype must_be_fresh: bool

    :ivar nonce: Nonce. ``None`` by default.
    :vartype nonce: int

    :ivar lifetime: InterestLifetime in milliseconds. ``4000`` by default.
    :vartype lifetime: int

    :ivar hop_limit: HopLimit. ``None`` by default.
    :vartype hop_limit: int

    :ivar forwarding_hint: ForwardingHint. The type should be list of pairs of Preference and Name.
        e.g.: ``[(1, "/ndn/name1"), (2, ["ndn", "name2"])]``
    :vartype forwarding_hint: :class:`List` [ :class:`Tuple` [ :class:`int` , :any:`NonStrictName` ]]
    """
    can_be_prefix: bool = False
    must_be_fresh: bool = False
    nonce: Optional[int] = None
    lifetime: Optional[int] = 4000
    hop_limit: Optional[int] = None
    forwarding_hint: List[Tuple[int, NonStrictName]] = dc.field(default_factory=list)

    @staticmethod
    def from_dict(kwargs):
        return InterestParam(**{f.name: kwargs[f.name]
                                for f in dc.fields(InterestParam)
                                if f.name in kwargs})


@dc.dataclass
class SignaturePtrs:
    r"""
    A set of pointers used to verify a packet.

    :ivar signature_info: the SignatureInfo.
    :vartype signature_info: :any:`SignatureInfo`

    :ivar signature_covered_part: a list of pointers, each of which points to a memory covered by signature.
    :vartype signature_covered_part: :class:`List` [ :class:`memoryview` ]

    :ivar signature_value_buf: a pointer to SignatureValue (TL excluded).
    :vartype signature_value_buf: :class:`memoryview`

    :ivar digest_covered_part: a list of pointers, each of which points to a memory covered by
        ParametersSha256DigestComponent.
    :vartype digest_covered_part: :class:`List` [ :class:`memoryview` ]

    :ivar digest_value_buf: a pointer to ParametersSha256DigestComponent (TL excluded).
    :vartype digest_value_buf: :class:`memoryview`
    """
    signature_info: Optional[SignatureInfo] = None
    signature_covered_part: Optional[List[BinaryStr]] = dc.field(default_factory=list)
    signature_value_buf: Optional[BinaryStr] = None
    digest_covered_part: Optional[List[BinaryStr]] = dc.field(default_factory=list)
    digest_value_buf: Optional[BinaryStr] = None


Interest = Tuple[FormalName, InterestParam, Optional[BinaryStr], SignaturePtrs]
Data = Tuple[FormalName, MetaInfo, Optional[BinaryStr], SignaturePtrs]


def make_interest(name: NonStrictName,
                  interest_param: InterestParam,
                  app_param: Optional[BinaryStr] = None,
                  signer: Optional[Signer] = None,
                  need_final_name: bool = False):
    r"""
    Make an Interest packet.

    :param name: the Name field.
    :type name: :any:`NonStrictName`
    :param interest_param: basic parameters of the Interest.
    :param app_param: the ApplicationParameters field.
    :type app_param: :class:`Optional` [ :any:`BinaryStr` ]
    :param signer: a Signer to sign this Interest. ``None`` if it is unsigned.
    :param need_final_name: if ``True``, also return the final Name with ParametersSha256DigestComponent.
    :return: TLV encoded Interest packet. If ``need_final_name``, return a tuple of the packet
        and the final Name.
    """
    interest = InterestPacket()
    interest.interest = InterestPacketValue()
    interest.interest.name = name
    interest.interest.can_be_prefix = interest_param.can_be_prefix
    interest.interest.must_be_fresh = interest_param.must_be_fresh
    interest.interest.nonce = interest_param.nonce
    interest.interest.lifetime = interest_param.lifetime
    interest.interest.hop_limit = interest_param.hop_limit

    if interest_param.forwarding_hint:
        interest.interest.forwarding_hint = Links()
        for preference, delegation in interest_param.forwarding_hint:
            cur = Delegation()
            cur.preference = preference
            cur.delegation = delegation
            interest.interest.forwarding_hint.delegations.append(cur)

    interest.interest.application_parameters = app_param
    if signer is not None:
        interest.interest.signature_info = SignatureInfo()
    markers = {}
    interest._signer.set_arg(markers, signer)
    ret = interest.encode(markers=markers)
    shrink_size = interest.interest._shrink_len.get_arg(markers['interest##inner_markers'])
    if shrink_size > 0:
        ret = shrink_length(ret, shrink_size)
    if need_final_name:
        return ret, InterestPacketValue.name.get_final_name(markers['interest##inner_markers'])
    else:
        return ret


def make_data(name: NonStrictName,
              meta_info: MetaInfo,
              content: Optional[BinaryStr] = None,
              signer: Optional[Signer] = None) -> VarBinaryStr:
    r"""
    Make a Data packet.

    :param name: the Name field.
    :type name: :any:`NonStrictName`
    :param meta_info: the MetaIndo field.
    :param content: the Content.
    :type content: :class:`Optional` [ :any:`BinaryStr` ]
    :param signer: a Signer to sign this Interest. ``None`` if it is unsigned.
    :return: TLV encoded Data packet.
    """
    data = DataPacket()
    data.data = DataPacketValue()
    data.data.meta_info = meta_info
    data.data.name = name
    data.data.content = content
    if signer is not None:
        data.data.signature_info = SignatureInfo()
    markers = {}
    data._signer.set_arg(markers, signer)
    ret = data.encode(markers=markers)
    shrink_size = data.data._shrink_len.get_arg(markers['data##inner_markers'])
    if shrink_size > 0:
        ret = shrink_length(ret, shrink_size)
    return ret


def parse_interest(wire: BinaryStr, with_tl: bool = True) -> Interest:
    r"""
    Parse a TLV encoded Interest.

    :param wire: the buffer.
    :type wire: :any:`BinaryStr`
    :param with_tl: ``True`` if the packet has Type and Length.
        ``False`` if ``wire`` only has the Value part.
    :return: a Tuple of Name, InterestParameters, ApplicationParameters and :any:`SignaturePtrs`.
    :rtype: :class:`Tuple` [ :any:`FormalName` , :any:`InterestParam` ,
        :class:`Optional` [ :any:`BinaryStr` ], :any:`SignaturePtrs` ]
    """
    if with_tl:
        wire = parse_and_check_tl(wire, TypeNumber.INTEREST)
    markers = {}
    ret = InterestPacketValue.parse(wire, markers)
    params = InterestParam()
    params.can_be_prefix = ret.can_be_prefix
    params.must_be_fresh = ret.must_be_fresh
    params.nonce = ret.nonce
    params.lifetime = ret.lifetime
    params.hop_limit = ret.hop_limit

    if ret.forwarding_hint and ret.forwarding_hint.delegations:
        for cur in ret.forwarding_hint.delegations:
            params.forwarding_hint.append((cur.preference, cur.delegation))

    sig_ptrs = SignaturePtrs(
        signature_info=ret.signature_info,
        signature_covered_part=ret._sig_cover_part.get_arg(markers),
        signature_value_buf=ret.signature_value,
        digest_covered_part=ret._digest_cover_part.get_arg(markers),
        digest_value_buf=ret._digest_buf.get_arg(markers)
    )
    return ret.name, params, ret.application_parameters, sig_ptrs


def parse_data(wire: BinaryStr, with_tl: bool = True) -> Data:
    r"""
    Parse a TLV encoded Data.

    :param wire: the buffer.
    :type wire: :any:`BinaryStr`
    :param with_tl: ``True`` if the packet has Type and Length.
        ``False`` if ``wire`` only has the Value part.
    :return: a Tuple of Name, MetaInfo, Content and :any:`SignaturePtrs`.
    :rtype: :class:`Tuple` [ :any:`FormalName` , :any:`MetaInfo` ,
        :class:`Optional` [ :any:`BinaryStr` ], :any:`SignaturePtrs` ]
    """
    if with_tl:
        wire = parse_and_check_tl(wire, TypeNumber.DATA)
    markers = {}
    ret = DataPacketValue.parse(wire, markers)
    params = ret.meta_info
    sig_ptrs = SignaturePtrs(
        signature_info=ret.signature_info,
        signature_covered_part=ret._sig_cover_part.get_arg(markers),
        signature_value_buf=ret.signature_value,
    )
    return ret.name, params, ret.content, sig_ptrs



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Optional
from .tlv_type import BinaryStr, VarBinaryStr
from .tlv_var import parse_and_check_tl
from .tlv_model import TlvModel, UintField, BytesField, ModelField, BoolField, DecodeError

__all__ = ['LpTypeNumber', 'NackReason', 'parse_network_nack', 'make_network_nack', 'parse_lp_packet']


class LpTypeNumber:
    FRAGMENT = 0x50
    SEQUENCE = 0x51
    FRAG_INDEX = 0x52
    FRAG_COUNT = 0x53
    HOP_COUNT = 0x54
    PIT_TOKEN = 0x62
    LP_PACKET = 0x64
    NACK = 0x0320
    NACK_REASON = 0x0321
    NEXT_HOP_FACE_ID = 0x0330
    INCOMING_FACE_ID = 0x0331
    CACHE_POLICY = 0x0334
    CACHE_POLICY_TYPE = 0x0335
    CONGESTION_MARK = 0x0340
    ACK = 0x0344
    TX_SEQUENCE = 0x0348
    NON_DISCOVERY = 0x034C
    PREFIX_ANNOUNCEMENT = 0x0350


class NackReason:
    NONE = 0
    CONGESTION = 50
    DUPLICATE = 100
    NO_ROUTE = 150


class NetworkNack(TlvModel):
    nack_reason = UintField(LpTypeNumber.NACK_REASON)


class CachePolicy(TlvModel):
    cache_policy_type = UintField(LpTypeNumber.CACHE_POLICY_TYPE)


class LpPacketValue(TlvModel):
    frag_index = UintField(LpTypeNumber.FRAG_INDEX)
    frag_count = UintField(LpTypeNumber.FRAG_COUNT)
    pit_token = BytesField(LpTypeNumber.PIT_TOKEN)
    nack = ModelField(LpTypeNumber.NACK, NetworkNack)
    next_hop_face_id = UintField(LpTypeNumber.NEXT_HOP_FACE_ID)
    incoming_face_id = UintField(LpTypeNumber.INCOMING_FACE_ID)
    cache_policy = ModelField(LpTypeNumber.CACHE_POLICY, CachePolicy)
    congestion_mark = UintField(LpTypeNumber.CONGESTION_MARK)
    tx_sequence = BytesField(LpTypeNumber.TX_SEQUENCE)
    ack = BytesField(LpTypeNumber.ACK)
    non_discovery = BoolField(LpTypeNumber.NON_DISCOVERY)
    prefix_announcement = BytesField(LpTypeNumber.PREFIX_ANNOUNCEMENT)

    fragment = BytesField(LpTypeNumber.FRAGMENT)


class LpPacket(TlvModel):
    lp_packet = ModelField(LpTypeNumber.LP_PACKET, LpPacketValue)


def parse_lp_packet(wire: BinaryStr, with_tl: bool = True) -> (Optional[int], Optional[BinaryStr]):
    """
    Parse an LpPacket, return NackReason (if exists) and the fragment.

    :param wire: an LpPacket.
    :param with_tl: if `wire` has the TL fields.
    :return: a tuple of NackReason and Fragment.
    """
    if with_tl:
        wire = parse_and_check_tl(wire, LpTypeNumber.LP_PACKET)
    markers = {}
    ret = LpPacketValue.parse(wire, markers, ignore_critical=True)

    if ret.frag_index is not None or ret.frag_count is not None:
        raise DecodeError('NDNLP fragmentation is not implemented yet.')

    if ret.nack is not None:
        return ret.nack.nack_reason, ret.fragment
    else:
        return None, ret.fragment


def parse_network_nack(wire: BinaryStr, with_tl: bool = True) -> (Optional[int], Optional[BinaryStr]):
    if with_tl:
        wire = parse_and_check_tl(wire, LpTypeNumber.LP_PACKET)
    markers = {}
    ret = LpPacketValue.parse(wire, markers, ignore_critical=True)

    if ret.nack is not None:
        return ret.nack.nack_reason, ret.fragment
    else:
        return None, None


def make_network_nack(encoded_interest: BinaryStr, nack_reason: int) -> VarBinaryStr:
    lp_packet = LpPacket()
    lp_packet.lp_packet = LpPacketValue()
    lp_packet.lp_packet.nack = NetworkNack()
    lp_packet.lp_packet.nack.nack_reason = nack_reason
    lp_packet.lp_packet.fragment = encoded_interest
    return lp_packet.encode()



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import abc
from typing import Tuple
from Cryptodome.Hash import SHA256
from Cryptodome.Random import get_random_bytes
from ...app_support.security_v2 import KEY_COMPONENT
from ...encoding import Signer, NonStrictName, BinaryStr, FormalName, Component


class Tpm(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_signer(self, key_name: NonStrictName) -> Signer:
        pass

    @abc.abstractmethod
    def generate_key(self, id_name: FormalName, key_type: str = 'rsa', **kwargs) -> Tuple[FormalName, BinaryStr]:
        pass

    @abc.abstractmethod
    def key_exist(self, key_name: FormalName) -> bool:
        pass

    @abc.abstractmethod
    def delete_key(self, key_name: FormalName):
        pass

    def construct_key_name(self, id_name: FormalName, pub_key: BinaryStr, **kwargs) -> FormalName:
        key_id = kwargs.pop('key_id', None)
        key_id_type = kwargs.pop('key_id_type', 'random')
        if not key_id:
            if key_id_type == 'random':
                while True:
                    key_id = Component.from_bytes(get_random_bytes(8))
                    if not self.key_exist(id_name + [KEY_COMPONENT, key_id]):
                        break
            elif key_id_type == 'sha256':
                h = SHA256.new()
                h.update(pub_key)
                key_id = Component.from_bytes(h.digest())
            else:
                raise ValueError(f'KeyIdType not supported: {key_id_type}')
        elif isinstance(key_id, str):
            key_id = Component.from_str(key_id)
        return id_name + [KEY_COMPONENT, key_id]



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import List, Union
from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import ECC
from Cryptodome.Signature import DSS
from ...encoding import Signer, SignatureType, KeyLocator, NonStrictName, VarBinaryStr


class Sha256WithEcdsaSigner(Signer):
    # SHA256 doesn't work for P-384 and P-521
    key_name: NonStrictName
    key_der: bytes
    curve_bit: int
    key_size: int

    def __init__(self, key_name: NonStrictName, key_der: Union[bytes, str]):
        self.key_name = key_name
        self.key_der = key_der
        self.key = ECC.import_key(self.key_der)
        curve = self.key.curve
        if curve[-2:] == 'r1' or curve[-2:] == 'v1':
            self.curve_bit = int(curve[-5:-2])
        else:
            self.curve_bit = int(curve[-3:])
        self.key_size = (self.curve_bit * 2 + 7) // 8
        self.key_size += self.key_size % 2

    def write_signature_info(self, signature_info):
        signature_info.signature_type = SignatureType.SHA256_WITH_ECDSA
        signature_info.key_locator = KeyLocator()
        signature_info.key_locator.name = self.key_name

    def get_signature_value_size(self):
        return self.key_size + 8

    def write_signature_value(self, wire: VarBinaryStr, contents: List[VarBinaryStr]) -> int:
        h = SHA256.new()
        for blk in contents:
            h.update(blk)
        signature = DSS.new(self.key, 'fips-186-3', 'der').sign(h)
        real_len = len(signature)
        wire[:real_len] = signature
        return real_len



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import abc
from typing import Optional
from ..encoding import SignaturePtrs, FormalName, InterestParam, BinaryStr
from ..encoding.signer import Signer
from ..types import Validator


class Policy:
    """
    Policy is an annotation attached to a node.
    """
    def __init__(self):
        self.node = None


class Cache(Policy, metaclass=abc.ABCMeta):
    """
    Cache policy determines how Data packets are stored.
    """
    @abc.abstractmethod
    async def search(self, match, name: FormalName, param: InterestParam):
        pass

    @abc.abstractmethod
    async def save(self, match, name: FormalName, packet: BinaryStr):
        pass


class InterestValidator(Policy, metaclass=abc.ABCMeta):
    """
    InterestValidator policy describes how to verify an Interest packet.
    """
    @abc.abstractmethod
    async def validate(self, match, sig_ptrs: SignaturePtrs) -> bool:
        pass


class DataValidator(Policy, metaclass=abc.ABCMeta):
    """
    DataValidator policy describes how to verify a Data packet.
    """
    @abc.abstractmethod
    def get_validator(self, match) -> Validator:
        pass


class Signing(Policy, metaclass=abc.ABCMeta):
    """
    Signing policy gives a signer used to sign a packet.
    When a user uses signing policy, he needs to specify whether its
    :class:`InterestSigning` or :class:`DataSigning`.
    """
    @abc.abstractmethod
    async def get_signer(self, match) -> Signer:
        pass


class InterestSigning(Signing, metaclass=abc.ABCMeta):
    """
    InterestSigning policy is a type used to indicate the Interest signer.
    Used as the type argument of set_policy.
    """
    pass


class DataSigning(Signing, metaclass=abc.ABCMeta):
    """
    DataSigning policy is a type used to indicate the Data signer.
    Used as the type argument of set_policy.
    """
    pass


class Encryption(Policy, metaclass=abc.ABCMeta):
    """
    Encryption policy encrypts and decrypts content.
    When a user uses encryption policy, he needs to specify whether its
    :class:`InterestEncryption` or :class:`DataEncryption`.
    """
    @abc.abstractmethod
    async def decrypt(self, match, content: BinaryStr) -> Optional[BinaryStr]:
        pass

    @abc.abstractmethod
    async def encrypt(self, match, content: BinaryStr) -> Optional[BinaryStr]:
        pass


class InterestEncryption(Encryption, metaclass=abc.ABCMeta):
    """
    InterestSigning policy is a type used to indicate the Interest encryption policy.
    Used as the type argument of set_policy.
    """
    pass


class DataEncryption(Encryption, metaclass=abc.ABCMeta):
    """
    DataEncryption policy is a type used to indicate the Data encryption policy.
    Used as the type argument of set_policy.
    """
    pass


class LocalOnly(Policy):
    """
    LocalOnly means the Data should be stored in the local storage.
    It prevents the node from sending Interest packets.
    """
    pass


class Register(Policy):
    """
    Register policy indicates the node should be registered as a prefix in the forwarder.
    """
    pass



from .tlv_type import *
from .tlv_var import *
from .name import *
from .signer import *
from .tlv_model import *

from .ndn_format_0_3 import *
from .ndnlp_v2 import *

__all__ = []
__all__.extend(tlv_type.__all__)
__all__.extend(tlv_var.__all__)
__all__.extend(name.__all__)
__all__.extend(signer.__all__)
__all__.extend(tlv_model.__all__)

__all__.extend(ndn_format_0_3.__all__)
__all__.extend(ndnlp_v2.__all__)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import logging
from typing import Callable, Dict, Any
from Cryptodome.PublicKey import ECC, RSA
from Cryptodome.Signature import DSS, pkcs1_15
from Cryptodome.Hash import SHA256
from ..encoding import SignaturePtrs, FormalName, Name, SignatureType
from ..types import Validator, NetworkError, InterestTimeout, InterestNack, ValidationFailure
from .schema_tree import Node
from . import policy


Checker = Callable[[Dict[str, Any], Dict[str, Any]], bool]


class SignedBy(policy.DataValidator, policy.InterestValidator):
    r"""
    SignedBy policy represents the trust schema,
    specifying the key used to signed the Interest or Data packet.
    It does the follows:

    - Match the key used to sign the packet in the static tree.
      The real key must match the node specified by ``key``.
      Otherwise, the validation fails.
    - Call the checker ``subject_to`` with two matching variable dict.
      Fail if the checker returns ``False``.
    - Call the ``need`` function of the matched key node to get the public key.
      Fail if the key cannot be fetched.
    - Verify the signature.

    .. note::

        Theoretically, SignedBy should also give the signer used to sign outgoing packets.
        However, this function is missing in current implementation.

    For example,

    .. code-block:: python3

        # This checker checks the Author of Data is the same as the Author of the key.
        def check_author(data_env, key_env):
            return data_env['Author'] == key_env['Author']

        root = Node()
        root['/author/<Author>/KEY/<KeyID>/self/<CertID>'] = Node()
        root['/blog/<Author>/<Category>/<Date>'] = Node()
        # The Data "/blog/<Author>/<Category>/<Date>" should be signed by
        # the key "/author/<Author>/KEY/<KeyID>" with the same author.
        root['/blog/<Author>/<Category>/<Date>'].set_policy(
            policy.DataValidator,
            SignedBy(root['/author/<Author>/KEY/<KeyID>'], subject_to=check_author))
    """
    def __init__(self, key: Node, subject_to: Checker = None):
        super().__init__()
        self.key = key
        self.subject_to = subject_to

    def get_validator(self, match) -> Validator:
        def validator(name: FormalName, sig_ptrs: SignaturePtrs):
            return self.validate(match.finer_match(name), sig_ptrs)
        return validator

    async def validate(self, match, sig_ptrs: SignaturePtrs) -> bool:
        # Check key name
        if sig_ptrs.signature_info is None or sig_ptrs.signature_info.key_locator is None:
            logging.info(f'{Name.to_str(match.name)} => Not signed')
            return False
        key_name = sig_ptrs.signature_info.key_locator.name
        if not key_name:
            logging.info(f'{Name.to_str(match.name)} => Not signed')
            return False
        key_match = match.root.match(key_name)
        if key_match.node is not self.key:
            logging.info(f'{Name.to_str(match.name)} => The key name {Name.to_str(key_name)} mismatch')
            return False
        if self.subject_to and not self.subject_to(match.env, key_match.env):
            logging.info(f'{Name.to_str(match.name)} => The key name {Name.to_str(key_name)} mismatch')
            return False
        # Get key_bits
        try:
            key_bits, _ = await key_match.need(must_be_fresh=True, can_be_prefix=True)
        except (NetworkError, InterestNack, InterestTimeout) as e:
            logging.info(f'{Name.to_str(match.name)} => Unable to fetch the key {Name.to_str(key_name)} due to {e}')
            return False
        except ValidationFailure:
            logging.info(f'{Name.to_str(match.name)} => The key {Name.to_str(key_name)} cannot be verified')
            return False
        # Import key
        sig_type = sig_ptrs.signature_info.signature_type
        key_bits = bytes(key_bits)
        try:
            if sig_type == SignatureType.SHA256_WITH_RSA:
                pub_key = RSA.import_key(key_bits)
                verifier = pkcs1_15.new(pub_key)
            elif sig_type == SignatureType.SHA256_WITH_ECDSA:
                pub_key = ECC.import_key(key_bits)
                verifier = DSS.new(pub_key, 'fips-186-3', 'der')
            else:
                logging.info(f'{Name.to_str(match.name)} => Unrecognized signature type {sig_type}')
                return False
        except (ValueError, IndexError, TypeError):
            logging.info(f'{Name.to_str(match.name)} => The key {Name.to_str(key_name)} is malformed')
            return False
        # Verify signature
        h = SHA256.new()
        for content in sig_ptrs.signature_covered_part:
            h.update(content)
        try:
            verifier.verify(h, bytes(sig_ptrs.signature_value_buf))
        except ValueError:
            logging.info(f'{Name.to_str(match.name)} => Unable to verify the signature')
            return False
        logging.debug(f'{Name.to_str(match.name)} => Verification passed')
        return True



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Union, List, Tuple
from ..encoding import Name, Component, BinaryStr

NamePattern = List[Union[BinaryStr, Tuple[int, int, str]]]
r"""
NamePattern is a list containing mixed name components and varaible patterns.
A variable pattern is a capturing pattern that matches with exactly one name component.
It is a tuple containing 3 variables:

- The 1st element is reserved and always 0. This is a quick and dirty solution in this PoC implementation
  It will be used if we want to support multiple name components matching patterns.
- The 2nd element is the TLV type of the name component to be matched.
- The 3rd element is the name of the pattern variable.
"""


def norm_pattern(name: str) -> NamePattern:
    """
    This function returns a normalized name pattern from a string, just like normalizing a name.

    :param name: the name pattern string.
    :return: normalized name pattern.
    """
    ret = Name.normalize(name)[:]
    for i, comp in enumerate(ret):
        comp_type = Component.get_type(comp)
        comp_value = Component.get_value(comp)
        if comp_type == Component.TYPE_GENERIC and comp_value[0] == b'<'[0] and comp_value[-1] == b'>'[0]:
            content = bytes(comp_value[1:-1]).decode()
            eq_sgn = content.find(':')
            if eq_sgn >= 0:
                type_str = content[:eq_sgn]
                if type_str == 'sha256digest':
                    type_val = Component.TYPE_IMPLICIT_SHA256
                elif type_str == 'params-sha256':
                    type_val = Component.TYPE_PARAMETERS_SHA256
                elif type_str in Component.ALTERNATE_URI_STR:
                    type_val = Component.ALTERNATE_URI_STR[type_str]
                else:
                    type_val = int(type_str)
                content = content[eq_sgn+1:]
            else:
                type_val = Component.TYPE_GENERIC
            if not content:
                raise ValueError('Pattern variable name cannot be empty')
            ret[i] = (0, type_val, content)
    return ret



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import abc
import struct
from typing import Optional, Type, List, Iterable
from functools import reduce
from .tlv_type import BinaryStr, VarBinaryStr, is_binary_str
from .tlv_var import write_tl_num, parse_tl_num, get_tl_num_size
from .name import Name, Component


__all__ = ['DecodeError', 'TlvModel', 'ProcedureArgument', 'OffsetMarker', 'UintField', 'BoolField',
           'NameField', 'BytesField', 'ModelField', 'RepeatedField', 'IncludeBase', 'IncludeBaseError']


class DecodeError(Exception):
    """
    Raised when there is a critical field (Type is odd) that is unrecognized, redundant or out-of-order.
    """
    pass


class IncludeBaseError(Exception):
    """
    Raised when IncludeBase is used to include a non-base class.
    """
    pass


class IncludeBase:
    """
    Include all fields from a base class.
    """
    def __init__(self, base):
        self.base = base


class TlvModelMeta(abc.ABCMeta):
    """
    Metaclass for TlvModel, used to collect fields.
    """
    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)

        # Collect encoded fields
        cls._encoded_fields = []
        index_dict = {}
        for field_name in cls.__dict__:
            if not field_name.startswith('__'):
                field_obj = getattr(cls, field_name)
                if isinstance(field_obj, Field):
                    field_obj.name = field_name
                    if field_name not in index_dict:
                        cls._encoded_fields.append(field_obj)
                        index_dict[field_name] = len(cls._encoded_fields) - 1
                    else:
                        cls._encoded_fields[index_dict[field_name]] = field_obj
                elif isinstance(field_obj, IncludeBase):
                    if field_obj.base not in bases:
                        raise IncludeBaseError(f"{field_obj.base} is not one of {name}'s base classes")
                    if not issubclass(field_obj.base, TlvModel):
                        raise IncludeBaseError(f"{field_obj.base} is not a TlvModel")
                    for field in field_obj.base._encoded_fields:
                        if field.name not in index_dict:
                            cls._encoded_fields.append(field)
                            index_dict[field.name] = len(cls._encoded_fields) - 1
                        else:
                            cls._encoded_fields[index_dict[field.name]] = field

        return cls


class Field(metaclass=abc.ABCMeta):
    """
    Field of :class:`TlvModel`.
    A field with value ``None`` will be omitted in encoding TLV.
    There is no required field in a :class:`TlvModel`, i.e. any Field can be ``None``.

    :ivar name: The name of the field
    :vartype name: str

    :ivar type_num: The Type number used in TLV encoding
    :vartype type_num: int

    :ivar default: The default value used for parsing and encoding.

        - If this field is absent during parsing, ``default`` is used to fill in this field.
        - If this field is not explicitly assigned to None before encoding,
          ``default`` is used.
    """
    def __init__(self, type_num: int, default=None):
        """
        Initialize a TLV field.

        :param type_num: Type number.
        :param default: default value used for parsing and encoding.
        """
        self.name = None
        self.type_num = type_num
        self.default = default

    def __get__(self, instance, owner):
        """
        Get the value of this field in a specific instance.
        Simply call :meth:`get_value` if ``instance`` is not ``None``.

        :param instance: the instance that this field is being accessed through.
        :param owner: the owner class of this field.
        :return: the value of this field.
        """
        if instance is None:
            return self
        return self.get_value(instance)

    def __set__(self, instance, value):
        """
        Set the value of this field.

        :param instance: the instance whose field is being set.
        :param value: the new value.
        """
        instance.__dict__[self.name] = value

    def get_value(self, instance):
        """
        Get the value of this field in a specific instance.
        Most fields use ``instance.__dict__`` to access the value.

        :param instance: the instance that this field is being accessed through.
        :return: the value of this field.
        """
        return instance.__dict__.get(self.name, self.default)

    @abc.abstractmethod
    def encoded_length(self, val, markers: dict) -> int:
        r"""
        Preprocess value and get encoded length of this field.
        The function may use ``markers[f'{self.name}##encoded_length']`` to store the length with TL.
        Other marker variables starting with ``f'{self.name}##'`` may also be used.
        Generally, marker variables are only used to store temporary values and avoid duplicated calculation.
        One field should not access to another field's marker by its name.

        This function may also use other marker variables. However, in that case,
        this field must be unique in a TlvModel. Usage of marker variables should follow
        the name convention defined by specific TlvModel.

        :param val: value of this field
        :param markers: encoding marker variables
        :return: encoded length with TL.
            It is expected as the exact length when encoding this field.
            The only exception is ``SignatureValueField`` (invisible to application developer).
        """
        pass

    @abc.abstractmethod
    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        """
        Encode this field into wire. Must be called after :meth:`encoded_length`.

        :param val: value of this field
        :param markers: encoding marker variables
        :param wire: buffer to encode
        :param offset: offset of this field in wire
        :return: encoded length with TL.
            It is expected to be the same as :meth:`encoded_length` returns.
        """
        pass

    @abc.abstractmethod
    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        """
        Parse the value of this field from an encoded wire.

        :param instance: the instance to parse into.
        :param markers: encoding marker variables. Only used in special cases.
        :param wire: the TLV encoded wire.
        :param offset: the offset of this field's Value in ``wire``.
        :param length: the Length of this field's Value.
        :param offset_btl: the offset of this field's TLV.

            .. code-block:: python3

                assert offset == (offset_btl
                                + get_tl_num_size(self.type_num)
                                + get_tl_num_size(length))

        :return: the value.
        """
        pass

    def skipping_process(self, markers: dict, wire: BinaryStr, offset: int):
        """
        Called when this field does not occur in ``wire`` and thus be skipped.

        :param markers: encoding marker variables.
        :param wire: the TLV encoded wire.
        :param offset: the offset where this field should have been if it occurred.
        """
        pass


class ProcedureArgument(Field):
    """
    A marker variable used during encoding or parsing.
    It does not have a value.
    Instead, it provides a way to access a specific variable in ``markers``.
    """
    def __init__(self, default=None):
        super().__init__(-1, default)

    def encoded_length(self, val, markers: dict) -> int:
        return 0

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        return 0

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        pass

    def __get__(self, instance, owner):
        """
        :return: itself.
        """
        return self

    def __set__(self, instance, value):
        """
        This is not allowed and will raise a :class:`TypeError` if called.
        """
        raise TypeError('ProcedureArgument can only be set via set_arg()')

    def get_arg(self, markers: dict):
        """
        Get its value from ``markers``

        :param markers: the markers dict.
        :return: its value.
        """
        return markers.get(f'{self.name}##args', self.default)

    def set_arg(self, markers: dict, val):
        """
        Set its value in ``markers``.

        :param markers: the markers dict.
        :param val: the new value.
        """
        markers[f'{self.name}##args'] = val


class OffsetMarker(ProcedureArgument):
    """
    A marker variable that records its position in TLV wire in terms of offset.
    """
    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        self.set_arg(markers, offset)
        return 0

    def skipping_process(self, markers: dict, wire: BinaryStr, offset: int):
        self.set_arg(markers, offset)


class UintField(Field):
    """
    NonNegativeInteger field.

    Type: :class:`int`

    Its Length is 1, 2, 4 or 8 when present.

    :ivar fixed_len: the fixed value for Length if it's not ``None``.
        Only 1, 2, 4 and 8 are acceptable.
    :vartype fixed_len: int
    """
    def __init__(self, type_num: int, default=None, fixed_len: int = None):
        super().__init__(type_num, default)
        if fixed_len not in {None, 1, 2, 4, 8}:
            raise ValueError("Uint's length should be 1, 2, 4, 8 or None")
        self.fixed_len = fixed_len

    def encoded_length(self, val, markers: dict) -> int:
        if val is None:
            return 0
        else:
            if not isinstance(val, int) or val < 0:
                raise TypeError(f'{self.name}=f{val} is not a legal uint')
            tl_size = get_tl_num_size(self.type_num) + 1
            if self.fixed_len is not None:
                ret = self.fixed_len
            else:
                if val <= 0xFF:
                    ret = 1
                elif val <= 0xFFFF:
                    ret = 2
                elif val <= 0xFFFFFFFF:
                    ret = 4
                else:
                    ret = 8
            if val >= 0x100 ** ret:
                raise ValueError(f'{val} cannot be encoded into {ret} bytes')
            markers[f'{self.name}##encoded_length'] = ret
            return ret + tl_size

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val is None:
            return 0
        tl_size = get_tl_num_size(self.type_num) + 1
        length = markers[f'{self.name}##encoded_length']
        offset += write_tl_num(self.type_num, wire, offset)
        if length == 1:
            struct.pack_into('!BB', wire, offset, 1, val)
        elif length == 2:
            struct.pack_into('!BH', wire, offset, 2, val)
        elif length == 4:
            struct.pack_into('!BI', wire, offset, 4, val)
        else:
            struct.pack_into('!BQ', wire, offset, 8, val)
        return length + tl_size

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        if length == 1:
            return struct.unpack_from('!B', wire, offset)[0]
        elif length == 2:
            return struct.unpack_from('!H', wire, offset)[0]
        elif length == 4:
            return struct.unpack_from('!I', wire, offset)[0]
        elif length == 8:
            return struct.unpack_from('!Q', wire, offset)[0]
        else:
            raise ValueError("Uint's length should be 1, 2, 4 or 8")


class BoolField(Field):
    """
    Boolean field.

    Type: :class:`bool`

    Its Length is always 0.
    When present, its Value is ``True``.
    When absent, its Value is ``None``, which is equivalent to ``False``.

    .. note::
        The default value is always ``None``.
    """
    def encoded_length(self, val, markers: dict) -> int:
        tl_size = get_tl_num_size(self.type_num) + 1
        return tl_size if val else 0

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val:
            tl_size = get_tl_num_size(self.type_num) + 1
            offset += write_tl_num(self.type_num, wire, offset)
            wire[offset] = 0
            return tl_size
        else:
            return 0

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        return True


class SignatureValueField(Field):
    def __init__(self,
                 type_num: int,
                 signer: ProcedureArgument,
                 covered_part: ProcedureArgument,
                 starting_point: OffsetMarker,
                 value_buffer: ProcedureArgument,
                 shrink_len: ProcedureArgument):
        super().__init__(type_num)
        self.signer = signer
        self.covered_part = covered_part
        self.starting_point = starting_point
        self.value_buffer = value_buffer
        self.shrink_len = shrink_len

    def encoded_length(self, val, markers: dict) -> int:
        signer = self.signer.get_arg(markers)
        if signer is None:
            return 0
        else:
            sig_value_len = signer.get_signature_value_size()
            length = 1 + get_tl_num_size(sig_value_len) + sig_value_len
            markers[f'{self.name}##encoded_length'] = sig_value_len
            return length

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        signer = self.signer.get_arg(markers)
        if signer is None:
            return 0
        else:
            sig_cover_start = self.starting_point.get_arg(markers)
            if sig_cover_start is not None:
                sig_cover_part = self.covered_part.get_arg(markers)
                sig_cover_part.append(wire[sig_cover_start:offset])

            origin_offset = offset
            sig_value_len = markers[f'{self.name}##encoded_length']
            offset += write_tl_num(self.type_num, wire, offset)
            markers[f'{self.name}##wire_length'] = wire[offset:offset+1]
            offset += write_tl_num(sig_value_len, wire, offset)
            self.value_buffer.set_arg(markers, wire[offset:offset + sig_value_len])
            offset += sig_value_len
            return offset - origin_offset

    def calculate_signature(self, markers: dict):
        signer = self.signer.get_arg(markers)
        if signer is not None:
            sig_value_len = markers[f'{self.name}##encoded_length']
            real_len = signer.write_signature_value(self.value_buffer.get_arg(markers),
                                                    self.covered_part.get_arg(markers))
            self.shrink_len.set_arg(markers, sig_value_len - real_len)
            if real_len != sig_value_len:
                if sig_value_len >= 253:
                    raise ValueError(f'Long signatrue with flexible length is not supported: {sig_value_len} >= 253')
                markers[f'{self.name}##wire_length'][0] = real_len

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        sig_buffer = memoryview(wire)[offset:offset+length]
        self.value_buffer.set_arg(markers, sig_buffer)

        sig_cover_start = self.starting_point.get_arg(markers)
        if sig_cover_start is not None:
            sig_cover_part = self.covered_part.get_arg(markers)
            sig_cover_part.append(wire[sig_cover_start:offset_btl])

        return sig_buffer


class InterestNameField(Field):
    def __init__(self,
                 need_digest: ProcedureArgument,
                 signature_covered_part: ProcedureArgument,
                 digest_buffer: ProcedureArgument,
                 default=None):
        super().__init__(Name.TYPE_NAME, default)
        self.need_digest = need_digest
        self.sig_covered_part = signature_covered_part
        self.digest_buffer = digest_buffer

    def encoded_length(self, val, markers: dict) -> int:
        digest_pos = None
        need_digest = self.need_digest.get_arg(markers)
        name = val
        if is_binary_str(name):
            # Decode it if it's binary name
            # This makes appending the digest component easier
            name = Name.decode(name)[0]
        elif isinstance(name, str):
            name = Name.from_str(name)
        elif isinstance(name, Iterable):
            # clone to prevent the list being modified
            name = list(name)
        # From here on, name must be in List[Component, str]
        if not isinstance(name, list):
            raise TypeError('invalid type for name')
        # Check every component
        for i, comp in enumerate(name):
            # If it's string, encode it first
            if isinstance(comp, str):
                name[i] = Component.from_str(Component.escape_str(comp))
                comp = name[i]
            # And then check the type
            if is_binary_str(comp):
                typ = Component.get_type(comp)
                if typ == Component.TYPE_INVALID:
                    raise TypeError('invalid type for name component')
                elif typ == Component.TYPE_PARAMETERS_SHA256:
                    # Params Sha256 can occur at most once
                    if need_digest and digest_pos is None:
                        digest_pos = i
                    else:
                        raise ValueError('unnecessary ParametersSha256DigestComponent in name')
            else:
                raise TypeError('invalid type for name component')
        markers[f'{self.name}##digest_pos'] = digest_pos
        markers[f'{self.name}##preprocessed_name'] = name

        length = reduce(lambda x, y: x + len(y), name, 0)
        if need_digest and digest_pos is None:
            length += 34
        markers[f'{self.name}##encoded_length'] = length
        return 1 + get_tl_num_size(length) + length

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        origin_offset = offset
        name_len = markers[f'{self.name}##encoded_length']
        name = markers[f'{self.name}##preprocessed_name']
        digest_pos = markers[f'{self.name}##digest_pos']
        need_digest = self.need_digest.get_arg(markers)
        sig_cover_part = self.sig_covered_part.get_arg(markers)
        digest_buf = None

        offset += write_tl_num(self.type_num, wire, offset)
        offset += write_tl_num(name_len, wire, offset)
        cover_start = offset  # Signature covers the name
        for i, comp in enumerate(name):
            wire[offset:offset + len(comp)] = comp
            if i == digest_pos:
                # except the Digest component
                if offset > cover_start:
                    sig_cover_part.append(wire[cover_start:offset])
                digest_buf = wire[offset + 2:offset + 34]
                cover_start = offset + 34
            offset += len(comp)
        if offset > cover_start:
            sig_cover_part.append(wire[cover_start:offset])
        if need_digest and digest_pos is None:
            markers[f'{self.name}##preprocessed_name'].append(wire[offset:offset+34])
            # If digest component does not exist, append one
            offset += write_tl_num(Component.TYPE_PARAMETERS_SHA256, wire, offset)
            offset += write_tl_num(32, wire, offset)
            digest_buf = wire[offset:offset + 32]
            offset += 32

        if need_digest:
            self.digest_buffer.set_arg(markers, digest_buf)
        return offset - origin_offset

    def get_final_name(self, markers):
        return markers[f'{self.name}##preprocessed_name']

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        name = Name.decode(wire, offset_btl)[0]
        sig_cover_part = self.sig_covered_part.get_arg(markers)
        for ele in name:
            typ = Component.get_type(ele)
            if typ == Component.TYPE_PARAMETERS_SHA256:
                self.digest_buffer.set_arg(markers, Component.get_value(ele))
            else:
                sig_cover_part.append(ele)
        return name


class NameField(Field):
    """
    NDN Name field. Its Type is always :any:`Name.TYPE_NAME`.

    Type: :any:`NonStrictName`
    """
    def __init__(self, default=None):
        super().__init__(Name.TYPE_NAME, default)

    def encoded_length(self, val, markers: dict) -> int:
        if val is None:
            return 0
        name = val
        if isinstance(name, str):
            name = Name.from_str(name)
        elif not is_binary_str(name):
            if isinstance(name, Iterable):
                name = list(name)
                for i, comp in enumerate(name):
                    if isinstance(comp, str):
                        name[i] = Component.from_str(Component.escape_str(comp))
                    elif not is_binary_str(comp):
                        raise TypeError('invalid type for name component')
            else:
                raise TypeError('invalid type for name')

        if isinstance(name, list):
            ret = Name.encoded_length(name)
        else:
            ret = len(name)
        markers[f'{self.name}##preprocessed_name'] = name
        markers[f'{self.name}##encoded_length_with_tl'] = ret
        return ret

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val is None:
            return 0
        name = markers[f'{self.name}##preprocessed_name']
        name_len_with_tl = markers[f'{self.name}##encoded_length_with_tl']
        if isinstance(name, list):
            Name.encode(name, wire, offset)
        else:
            wire[offset:offset + name_len_with_tl] = name
        return name_len_with_tl

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        return Name.decode(wire, offset_btl)[0]


class BytesField(Field):
    r"""
    Field for ``*OCTET``.

    Type: :any:`BinaryStr`

    .. note::
        Do not assign it with a :class:`str`.
    """
    def encoded_length(self, val, markers: dict) -> int:
        if val is None:
            return 0
        tl_size = get_tl_num_size(self.type_num) + get_tl_num_size(len(val))
        return tl_size + len(val)

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val is None:
            return 0
        else:
            origin_offset = offset
            offset += write_tl_num(self.type_num, wire, offset)
            offset += write_tl_num(len(val), wire, offset)
            wire[offset:offset+len(val)] = val
            offset += len(val)
            return offset - origin_offset

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        return memoryview(wire)[offset:offset+length]


class TlvModel(metaclass=TlvModelMeta):
    r"""
    Used to describe a TLV format.

    :ivar _encoded_fields: a list of :any:`Field` in order.
    :vartype _encoded_fields: List[Field]
    """
    _encoded_fields: List[Field]

    def __repr__(self):
        values = ', '.join(f'{field.name}={field.get_value(self).__repr__()}' for field in self._encoded_fields)
        return f'{self.__class__.__name__}({values})'

    def __eq__(self, other):
        """
        Compare two TlvModels

        :param other: the other TlvModel to compare with.
        :return: whether all Fields are equal.
        """
        for field in self._encoded_fields:
            if field.get_value(self) != field.get_value(other):
                return False
        return True

    def asdict(self, dict_factory=dict):
        """
        Return a dict to represent this TlvModel.

        :param dict_factory: class of dict.
        :return: the dict.
        """
        result = []
        for field in self._encoded_fields:
            if isinstance(field, ModelField):
                result.append((field.name, field.get_value(self).asdict()))
            elif isinstance(field, RepeatedField):
                result.append((field.name, field.aslist(self)))
            elif isinstance(field, BytesField):
                result.append((field.name, bytes(field.get_value(self))))
            else:
                result.append((field.name, field.get_value(self)))
        return dict_factory(result)

    def encoded_length(self, markers: Optional[dict] = None) -> int:
        """
        Get the encoded Length of this TlvModel.

        :param markers: encoding marker variables.
        :return: the encoded Length.
        """
        if markers is None:
            markers = {}
        ret = 0
        for field in self._encoded_fields:
            ret += field.encoded_length(field.get_value(self), markers)
        markers[f'##encoded_length'] = ret
        return ret

    def encode(self,
               wire: VarBinaryStr = None,
               offset: int = 0,
               markers: Optional[dict] = None) -> VarBinaryStr:
        r"""
        Encode the TlvModel.

        :param wire: the buffer to contain the encoded wire.
            A new :class:`bytearray` will be created if it's ``None``.
        :param offset: the starting offset.
        :param markers: encoding marker variables.
        :return: wire.

        :raises ValueError: some field is assigned with improper value.
        :raises TypeError: some field is assigned with value of wrong type.
        :raises IndexError: wire does not have enough length.
        :raises struct.error: a negative number is assigned to any non-negative integer field.
        """
        if markers is None:
            markers = {}
        if f'##encoded_length' in markers:
            length = markers[f'##encoded_length']
        else:
            length = self.encoded_length(markers)
        if wire is None:
            wire = bytearray(length)
        wire_view = memoryview(wire)
        for field in self._encoded_fields:
            offset += field.encode_into(field.get_value(self), markers, wire_view, offset)
        return wire

    @classmethod
    def parse(cls, wire: BinaryStr, markers: Optional[dict] = None, ignore_critical: bool = False):
        """
        Parse a TlvModel from TLV encoded wire.

        :param wire: the TLV encoded wire.
        :param markers: encoding marker variables.
        :param ignore_critical: whether to ignore unknown critical fields.
        :return: parsed TlvModel.

        :raises DecodeError: a critical field is unrecognized, redundant or out-of-order.
        :raises IndexError: the Length of a field exceeds the size of wire.
        """
        if markers is None:
            markers = {}
        offset = 0
        field_pos = 0
        ret = cls()
        ret.__dict__ = {}  # Clean default values created in __init__
        while offset < len(wire):
            # Read TL
            offset_btl = offset
            typ, size_typ = parse_tl_num(wire, offset)
            offset += size_typ
            length, size_len = parse_tl_num(wire, offset)
            offset += size_len
            # Search for field
            i = field_pos
            while i < len(ret._encoded_fields):
                if ret._encoded_fields[i].type_num == typ:
                    break
                i += 1
            if i < len(ret._encoded_fields):
                # If found
                # First process skipped fields
                for j in range(field_pos, i):
                    ret._encoded_fields[j].skipping_process(markers, wire, offset_btl)
                # Parse that field
                val = ret._encoded_fields[i].parse_from(ret, markers, wire, offset, length, offset_btl)
                ret._encoded_fields[i].__set__(ret, val)
                # Set next field
                if isinstance(ret._encoded_fields[i], RepeatedField):
                    field_pos = i
                else:
                    field_pos = i + 1
            else:
                # If not found
                if (typ & 1) == 1 and not ignore_critical:
                    raise DecodeError(f'a critical field of type {typ} is unrecognized, redundant or out-of-order')
            offset += length
        return ret


class ModelField(Field):
    r"""
    Field for nested TlvModel.

    Type: :any:`TlvModel`

    :ivar model_type: the type of its value.
    :vartype model_type: :any:`TlvModelMeta`

    :ivar ignore_critical: whether to ignore critical fields (whose Types are odd).
    :vartype ignore_critical: :class:`bool`
    """
    def __init__(self,
                 type_num: int,
                 model_type: Type[TlvModel],
                 copy_in_fields: List[ProcedureArgument] = None,
                 copy_out_fields: List[ProcedureArgument] = None,
                 ignore_critical: bool = False):
        # default should be None here to prevent unintended modification
        super().__init__(type_num, None)
        self.model_type = model_type
        self.copy_in_fields = copy_in_fields if copy_in_fields else {}
        self.copy_out_fields = copy_out_fields if copy_out_fields else {}
        self.ignore_critical = ignore_critical

    def encoded_length(self, val, markers: dict) -> int:
        if val is None:
            return 0
        if not isinstance(val, self.model_type):
            raise TypeError(f'{self.name}=f{val} is of type {self.model_type}')
        copy_fields = {f.name for f in self.copy_in_fields}
        inner_markers = {k: v
                         for k, v in markers.items()
                         if k.split('##')[0] in copy_fields}
        length = val.encoded_length(inner_markers)
        markers[f'{self.name}##inner_markers'] = inner_markers
        markers[f'{self.name}##encoded_length'] = length
        return get_tl_num_size(self.type_num) + get_tl_num_size(length) + length

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val is None:
            return 0
        else:
            inner_markers = markers[f'{self.name}##inner_markers']
            length = markers[f'{self.name}##encoded_length']

            origin_offset = offset
            offset += write_tl_num(self.type_num, wire, offset)
            offset += write_tl_num(length, wire, offset)
            val.encode(wire, offset, inner_markers)
            offset += length
            return offset - origin_offset

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        inner_markers = {}
        val = self.model_type.parse(memoryview(wire)[offset:offset+length], inner_markers, self.ignore_critical)
        copy_fields = {f.name for f in self.copy_out_fields}
        for k, v in inner_markers.items():
            if k.split('##')[0] in copy_fields:
                markers[k] = v
        return val


class RepeatedField(Field):
    r"""
    Field for an array of a specific type.
    All elements will be directly encoded into TLV wire in order, sharing the same Type.
    The ``type_num`` of ``element_type`` is used.

    Type: :class:`list`

    :vartype element_type: :any:`Field`
    :ivar element_type: the type of elements in the list.

        .. warning::

            Please always create a new :any:`Field` instance.
            Don't use an existing one.
    """
    def __init__(self, element_type: Field):
        # default should be None here to prevent unintended modification
        super().__init__(element_type.type_num, None)
        self.element_type = element_type

    def get_value(self, instance):
        if self.name not in instance.__dict__:
            instance.__dict__[self.name] = []
        return instance.__dict__[self.name]

    def encoded_length(self, val, markers: dict) -> int:
        if not val:
            return 0

        ret = 0
        # Different from ModelField, here changing the name is allowed
        # Because self.element_type is always a new field instance
        # ModelField share a ModelClass with others, and also
        # subfields under a model do not use its name prefix so
        # there may be conflicts
        for i, ele in enumerate(val):
            self.element_type.name = f'{self.name}[{i}]'
            ret += self.element_type.encoded_length(ele, markers)

        return ret  # TL is not included here

    def encode_into(self, val, markers: dict, wire: VarBinaryStr, offset: int) -> int:
        if val is None:
            return 0
        else:
            origin_offset = offset
            for i, ele in enumerate(val):
                self.element_type.name = f'{self.name}[{i}]'
                offset += self.element_type.encode_into(ele, markers, wire, offset)
            return offset - origin_offset

    def parse_from(self, instance, markers: dict, wire: BinaryStr, offset: int, length: int, offset_btl: int):
        lst = self.get_value(instance)
        self.element_type.name = f'{self.name}[{len(lst)}]'
        new_ele = self.element_type.parse_from(instance, markers, wire, offset, length, offset_btl)
        lst.append(new_ele)
        return lst

    def aslist(self, instance):
        ret = []
        for x in self.get_value(instance):
            if isinstance(x, TlvModel):
                ret.append(x.asdict())
            elif isinstance(x, memoryview):
                ret.append(bytes(x))
            else:
                ret.append(x)
        return ret



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import io
import abc
import asyncio as aio
from typing import Optional, Callable, Coroutine, Any
from ..encoding.tlv_var import read_tl_num_from_stream


class Face(metaclass=abc.ABCMeta):
    running: bool = False
    callback: Callable[[int, bytes], Coroutine[Any, None, None]] = None

    def __init__(self):
        self.running = False

    @abc.abstractmethod
    async def open(self):
        pass

    @abc.abstractmethod
    def shutdown(self):
        pass

    @abc.abstractmethod
    def send(self, data: bytes):
        pass

    @abc.abstractmethod
    async def run(self):
        pass


class StreamFace(Face, metaclass=abc.ABCMeta):
    reader: Optional[aio.StreamReader] = None
    writer: Optional[aio.StreamWriter] = None

    def shutdown(self):
        self.running = False
        if self.writer:
            self.writer.close()
            self.writer = None

    async def run(self):
        while self.running:
            try:
                bio = io.BytesIO()
                typ = await read_tl_num_from_stream(self.reader, bio)
                siz = await read_tl_num_from_stream(self.reader, bio)
                bio.write(await self.reader.readexactly(siz))
                buf = bio.getvalue()
                aio.ensure_future(self.callback(typ, buf))
            except aio.IncompleteReadError:
                self.shutdown()

    def send(self, data: bytes):
        self.writer.write(data)


class UnixFace(StreamFace):
    path: str = '/run/nfd.sock'

    def __init__(self, path: str = ''):
        super().__init__()
        if path:
            self.path = path

    async def open(self):
        self.reader, self.writer = await aio.open_unix_connection(self.path)
        self.running = True


class TcpFace(StreamFace):
    host: str = '127.0.0.1'
    port: int = 6363

    def __init__(self, host: str = '', port: int = 0):
        super().__init__()
        if host:
            self.host = host
        if port:
            self.port = port

    async def open(self):
        self.reader, self.writer = await aio.open_connection(self.host, self.port)
        self.running = True



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import abc
from typing import List
from .tlv_type import VarBinaryStr


__all__ = ['Signer']


class Signer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def write_signature_info(self, signature_info):
        """
        Fill in the fields of SignatureInfo.

        :param signature_info: a blank SignatureInfo object.
        """
        pass

    @abc.abstractmethod
    def get_signature_value_size(self) -> int:
        """
        Get the size of SignatureValue.
        If the size is variable, return the maximum possible value.

        :return: the size of SignatureValue.
        """
        pass

    @abc.abstractmethod
    def write_signature_value(self, wire: VarBinaryStr, contents: List[VarBinaryStr]) -> int:
        """
        Calculate the SignatureValue and write it into wire.
        The length of wire is exactly what :meth:`get_signature_value_size` returns.
        Basically this function should return the same value except for ECDSA.

        :param wire: the buffer to contain SignatureValue.
        :param contents: a list of memory blocks that needs to be covered.
        :return: the actual size of SignatureValue.
        """
        pass



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
"""
Component module is a collection of functions processing NDN NameComponents.
In python-ndn, a NameComponent is always encoded in TLV form, of type :class:`bytes`,
:class:`bytearray` or :class:`memoryview`.

The types of NameComonent follows
`Name Component Assignment policy <https://redmine.named-data.net/projects/ndn-tlv/wiki/NameComponentType>`_.
Type constants are following

======================   =====================================
Type                     Description
======================   =====================================
TYPE_INVALID             Invalid name component type
TYPE_GENERIC             Implicit SHA-256 digest component
TYPE_IMPLICIT_SHA256     SHA-256 digest of Interest Parameters
TYPE_PARAMETERS_SHA256   Generic name component
TYPE_KEYWORD             Well-known keyword
TYPE_SEGMENT             Segment number
TYPE_BYTE_OFFSET         Byte offset
TYPE_VERSION             Version number
TYPE_TIMESTAMP           Unix timestamp in microseconds
TYPE_SEQUENCE_NUM        Sequence number
======================   =====================================

"""
import string
from ..tlv_type import BinaryStr
from ..tlv_var import write_tl_num, pack_uint_bytes, parse_tl_num, get_tl_num_size

CHARSET = (set(string.ascii_letters)
           | set(string.digits)
           | {'-', '.', '_', '~', '=', '%'})
"""The character set for NameComponent, which is unreserved characters + {'=', '%'}"""

TYPE_INVALID = 0x00
TYPE_GENERIC = 0x08
TYPE_IMPLICIT_SHA256 = 0x01
TYPE_PARAMETERS_SHA256 = 0x02
TYPE_KEYWORD = 0x20
TYPE_SEGMENT = 0x21
TYPE_BYTE_OFFSET = 0x22
TYPE_VERSION = 0x23
TYPE_TIMESTAMP = 0x24
TYPE_SEQUENCE_NUM = 0x25

ALTERNATE_URI_TYPE = {
    TYPE_SEGMENT: 'seg={}',
    TYPE_BYTE_OFFSET: 'off={}',
    TYPE_VERSION: 'v={}',
    TYPE_TIMESTAMP: 't={}',
    TYPE_SEQUENCE_NUM: 'seq={}'
}

ALTERNATE_URI_STR = {
    'seg': TYPE_SEGMENT,
    'off': TYPE_BYTE_OFFSET,
    'v': TYPE_VERSION,
    't': TYPE_TIMESTAMP,
    'seq': TYPE_SEQUENCE_NUM,
}


def from_bytes(val: BinaryStr, typ: int = TYPE_GENERIC) -> bytearray:
    """
    Construct a Component from bytes by adding a type and length.

    :param val: the value of the component.
    :param typ: the type of the component. :const:`TYPE_GENERIC` by default.
    :return: the component.
    """
    size_typ = get_tl_num_size(typ)
    size_len = get_tl_num_size(len(val))
    ret = bytearray(size_typ + size_len + len(val))
    write_tl_num(typ, ret, 0)
    write_tl_num(len(val), ret, size_typ)
    ret[size_typ+size_len:] = val
    return ret


def from_hex(val: str, typ: int = TYPE_GENERIC) -> bytearray:
    """
    Construct a Component from hex string.

    :param val: a hexadecimal string.
    :param typ: the type of the component. :const:`TYPE_GENERIC` by default.
    :return: the component.
    """
    return from_bytes(bytearray.fromhex(val), typ)


def from_str(val: str) -> bytearray:
    """
    Construct a Component from URI string.

    :param val: URI string. All characters should be from :const:`CHARSET`, otherwise it would
        raise a :obj:`ValueError`.

        .. note::
            Additional periods are not allowed here. To create a zero-size Component, just pass
            an empty string ``''`` in.

    :return: the component.
    :raises ValueError: the string is not a legal URI.
    """
    def raise_except():
        raise ValueError(f'{val} is not a legal Name.')

    # Check empty string
    if not val:
        return bytearray(b'\x08\x00')

    percent_cnt = 0
    type_offset = None
    # Check charset
    for i, ch in enumerate(val):
        if ch not in CHARSET:
            raise_except()
        if ch == '%':
            percent_cnt += 1
        if ch == "=":
            if type_offset is not None:
                raise_except()
            else:
                type_offset = i
    # Get Type
    typ = TYPE_GENERIC
    if type_offset is not None:
        try:
            typ_str = val[:type_offset]
            # Check special case
            if typ_str == 'sha256digest':
                return from_bytes(bytearray.fromhex(val[type_offset + 1:]), TYPE_IMPLICIT_SHA256)
            elif typ_str == 'params-sha256':
                return from_bytes(bytearray.fromhex(val[type_offset + 1:]), TYPE_PARAMETERS_SHA256)
            elif typ_str in ALTERNATE_URI_STR:
                return from_number(int(val[type_offset + 1:]), ALTERNATE_URI_STR[typ_str])
            # General case
            else:
                typ = int(typ_str)
        except ValueError:
            raise_except()
    else:
        typ = TYPE_GENERIC
        type_offset = -1
    # Alloc buf
    length = len(val) - type_offset - 1 - 2 * percent_cnt
    if length < 0:
        raise_except()
    size_typ = get_tl_num_size(typ)
    size_len = get_tl_num_size(length)
    ret = bytearray(size_typ + size_len + length)
    view = memoryview(ret)
    write_tl_num(typ, ret, 0)
    write_tl_num(length, ret, size_typ)

    # Encode val
    i = type_offset + 1
    pos = size_typ + size_len

    def encode():
        if val[i] != '%':
            view[pos] = val[i].encode('utf-8')[0]
            return 1
        else:
            view[pos] = int(val[i+1:i+3], 16)
            return 3

    while i < len(val):
        try:
            i += encode()
            pos += 1
        except IndexError:
            raise_except()
    return ret


def from_number(val: int, typ: int) -> bytearray:
    """
    Construct a Component from an integer.

    :param val: the integer.
    :param typ: the type of the component.
    :return: the component.
    """
    return from_bytes(pack_uint_bytes(val), typ)


def from_segment(segment: int) -> bytearray:
    """
    Construct a Component from an segment number.

    :param segment: the segment number.
    :return: the component.
    """
    return from_number(segment, TYPE_SEGMENT)


def from_byte_offset(offset: int) -> bytearray:
    """
    Construct a Component from a byte offset.

    :param offset: the byte offset.
    :return: the component.
    """
    return from_number(offset, TYPE_BYTE_OFFSET)


def from_sequence_num(seq_num: int) -> bytearray:
    """
    Construct a Component from a sequence number.

    :param seq_num: the sequence number.
    :return: the component.
    """
    return from_number(seq_num, TYPE_SEQUENCE_NUM)


def from_version(version: int) -> bytearray:
    """
    Construct a Component from a version number.

    :param version: the version number.
    :return: the component.
    """
    return from_number(version, TYPE_VERSION)


def from_timestamp(timestamp: int) -> bytearray:
    """
    Construct a Component from a timestamp number.

    :param timestamp: the timestamp
    :return: the component.

    :examples:
        >>> from ndn.encoding.name import Component
        >>> from ndn.utils import timestamp
        >>> Component.to_str(Component.from_timestamp(timestamp()))
        '36=%00%00%01nH.%A7%90'
    """
    return from_number(timestamp, TYPE_TIMESTAMP)


def get_type(component: BinaryStr) -> int:
    """
    Get the type from a Component.

    :param component: the component.
    :return: the type.
    """
    return parse_tl_num(component)[0]


def get_value(component: BinaryStr) -> memoryview:
    """
    Get the value from a Component, in the form of :class:`memoryview`.

    :param component: the component.
    :return: the value.
    """
    _, size_typ = parse_tl_num(component, 0)
    _, size_len = parse_tl_num(component, size_typ)
    return memoryview(component)[size_typ + size_len:]


def to_str(component: BinaryStr) -> str:
    """
    Convert a Component into a URI string. Returns an empty string ``''`` for a 0-size Component.

    :param component: the component.
    :return: a URI string.
    """
    offset = 0
    typ, sz = parse_tl_num(component, offset)
    offset += sz
    length, sz = parse_tl_num(component, offset)
    offset += sz
    if len(component) != length + offset:
        raise ValueError(f'{component} is malformed.')

    if typ == TYPE_IMPLICIT_SHA256:
        return f"sha256digest={component[offset:].hex()}"
    elif typ == TYPE_PARAMETERS_SHA256:
        return f"params-sha256={component[offset:].hex()}"
    elif typ in ALTERNATE_URI_TYPE:
        return ALTERNATE_URI_TYPE[typ].format(int.from_bytes(component[offset:], 'big'))
    else:
        ret = ""
        if typ != TYPE_GENERIC:
            ret = f"{typ}="

        def decode(val: int) -> str:
            ret = chr(val)
            if ret in CHARSET and ret not in {'%', '='}:
                return ret
            else:
                return f"%{val:02X}"

        return ret + "".join(decode(val) for val in component[offset:])


def to_number(component: BinaryStr) -> int:
    """
    Take the number encoded in the component out.

    :param component: the component.
    :return: an integer, which is the value of the component.
    """
    _, size_typ = parse_tl_num(component, 0)
    _, size_len = parse_tl_num(component, size_typ)
    return int.from_bytes(component[size_typ + size_len:], 'big')


def escape_str(val: str) -> str:
    r"""
    Escape a string to a legal URI string.
    Any characters not in the :const:`CHARSET` will be converted into percent-hexadecimal encoding.
    ``'%'`` itself will not be escaped. For hex digits, lowercase is used.

    :param val: the string to escape.
    :return: the URI string.

    :examples:
        >>> from ndn.encoding.name import Component
        >>> Component.escape_str('Kraus Bölter')
        'Kraus%20B%C3%B6lter'

        >>> Component.escape_str('all:%0a\tgcc -o a.out')
        'all%3A%0a%09gcc%20-o%20a.out'
    """
    def escape_chr(ch):
        if ch in CHARSET:
            return ch
        else:
            return ''.join(f'%{x:02X}' for x in ch.encode())

    return ''.join(escape_chr(ch) for ch in val)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import struct
import logging
import asyncio as aio
from typing import Optional, Any, Awaitable, Coroutine, Tuple, List
from .utils import gen_nonce
from .encoding import BinaryStr, TypeNumber, LpTypeNumber, parse_interest, \
    parse_tl_num, parse_data, DecodeError, Name, NonStrictName, MetaInfo, \
    make_data, InterestParam, make_interest, FormalName, SignaturePtrs, parse_lp_packet
from .security import Keychain, sha256_digest_checker, params_sha256_checker
from .transport.stream_socket import Face
from .app_support.nfd_mgmt import make_command, parse_response
from .name_tree import NameTrie, InterestTreeNode, PrefixTreeNode
from .types import NetworkError, InterestTimeout, Validator, Route, InterestCanceled, \
    InterestNack, ValidationFailure
from .client_conf import read_client_conf, default_face, default_keychain


class NDNApp:
    """
    An NDN application.

    :ivar face: the Face used to connection to a NFD node.
    :ivar keychain: the Keychain to store Identities and Keys, providing Signers.
    :ivar int_validator: the default validator for Interest packets.
    :ivar data_validator: the default validator for Data packets.
    """
    face: Face = None
    keychain: Keychain = None
    _int_tree: NameTrie = None
    _prefix_tree: NameTrie = None
    int_validator: Validator = None
    data_validator: Validator = None
    _autoreg_routes: List[Tuple[FormalName, Route, Optional[Validator], bool, bool]]
    _prefix_register_semaphore: aio.Semaphore = None

    def __init__(self, face=None, keychain=None):
        config = read_client_conf() if not face or not keychain else {}
        if face is not None:
            self.face = face
        else:
            self.face = default_face(config['transport'])
        self.face.callback = self._receive
        if keychain is not None:
            self.keychain = keychain
        else:
            self.keychain = default_keychain(config['pib'], config['tpm'])
        self._int_tree = NameTrie()
        self._prefix_tree = NameTrie()
        self.data_validator = sha256_digest_checker
        self.int_validator = sha256_digest_checker
        self._autoreg_routes = []
        self._prefix_register_semaphore = aio.Semaphore(1)

    async def _receive(self, typ: int, data: BinaryStr):
        """
        Pipeline when a packet is received.

        :param typ: the Type.
        :param data: the Value of the packet with TL.
        """
        logging.debug('Packet received %s, %s' % (typ, bytes(data)))
        if typ == LpTypeNumber.LP_PACKET:
            try:
                nack_reason, fragment = parse_lp_packet(data, with_tl=True)
            except (DecodeError, TypeError, ValueError, struct.error):
                logging.warning('Unable to decode received packet')
                return
            data = fragment
            typ, _ = parse_tl_num(data)
        else:
            nack_reason = None

        if nack_reason is not None:
            try:
                name, _, _, _ = parse_interest(data, with_tl=True)
            except (DecodeError, TypeError, ValueError, struct.error):
                logging.warning('Unable to decode the fragment of LpPacket')
                return
            logging.debug('NetworkNack received %s, reason=%s' % (Name.to_str(name), nack_reason))
            self._on_nack(name, nack_reason)
        else:
            if typ == TypeNumber.INTEREST:
                try:
                    name, param, app_param, sig = parse_interest(data, with_tl=True)
                except (DecodeError, TypeError, ValueError, struct.error):
                    logging.warning('Unable to decode received packet')
                    return
                logging.debug('Interest received %s' % Name.to_str(name))
                await self._on_interest(name, param, app_param, sig, raw_packet=data)
            elif typ == TypeNumber.DATA:
                try:
                    name, meta_info, content, sig = parse_data(data, with_tl=True)
                except (DecodeError, TypeError, ValueError, struct.error):
                    logging.warning('Unable to decode received packet')
                    return
                logging.debug('Data received %s' % Name.to_str(name))
                await self._on_data(name, meta_info, content, sig, raw_packet=data)
            else:
                logging.warning('Unable to decode received packet')

    def put_raw_packet(self, data: BinaryStr):
        r"""
        Send a raw Data packet.

        :param data: TLV encoded Data packet.
        :type data: :any:`BinaryStr`
        :raises NetworkError: the face to NFD is down.
        """
        if not self.face.running:
            raise NetworkError('cannot send packet before connected')
        self.face.send(data)

    def prepare_data(self, name: NonStrictName, content: Optional[BinaryStr] = None, **kwargs):
        r"""
        Prepare a Data packet by generating, encoding and signing it.

        :param name: the Name.
        :type name: :any:`NonStrictName`
        :param content: the Content.
        :type content: Optional[:any:`BinaryStr`]
        :param kwargs: :ref:`label-keyword-arguments`.
        :return: TLV encoded Data packet.
        """
        if 'signer' in kwargs:
            signer = kwargs['signer']
        else:
            signer = self.keychain.get_signer(kwargs)
        if 'meta_info' in kwargs:
            meta_info = kwargs['meta_info']
        else:
            meta_info = MetaInfo.from_dict(kwargs)
        return make_data(name, meta_info, content, signer=signer)

    def put_data(self, name: NonStrictName, content: Optional[BinaryStr] = None, **kwargs):
        r"""
        Publish a Data packet.

        :param name: the Name.
        :type name: :any:`NonStrictName`
        :param content: the Content.
        :type content: Optional[:any:`BinaryStr`]
        :param kwargs: :ref:`label-keyword-arguments`.
        :return: TLV encoded Data packet.
        """
        self.put_raw_packet(self.prepare_data(name, content, **kwargs))

    def express_interest(self,
                         name: NonStrictName,
                         app_param: Optional[BinaryStr] = None,
                         validator: Optional[Validator] = None,
                         need_raw_packet: bool = False,
                         **kwargs) -> Coroutine[Any, None, Tuple[FormalName, MetaInfo, Optional[BinaryStr]]]:
        r"""
        Express an Interest packet.

        The Interest packet is sent immediately and a coroutine used to get the result is returned.
        Awaiting on what is returned will block until the Data is received and return that Data.
        An exception is raised if unable to receive the Data.

        :param name: the Name.
        :type name: :any:`NonStrictName`
        :param app_param: the ApplicationParameters.
        :type app_param: Optional[:any:`BinaryStr`]
        :param validator: the Validator used to verify the Data received.
        :type validator: Optional[:any:`Validator`]
        :param need_raw_packet: if True, return the raw Data packet with TL.
        :type need_raw_packet: bool
        :param kwargs: :ref:`label-keyword-arguments`.
        :return: A tuple of (Name, MetaInfo, Content) after ``await``.
            If need_raw_packet is True, return a tuple (Name, MetaInfo, Content, RawPacket).
        :rtype: Coroutine[Any, None, Tuple[:any:`FormalName`, :any:`MetaInfo`, Optional[:any:`BinaryStr`]]]

        The following exception is raised by ``express_interest``:

        :raises NetworkError: the face to NFD is down before sending this Interest.

        The following exceptions are raised by the coroutine returned:

        :raises InterestNack: an NetworkNack is received.
        :raises InterestTimeout: time out.
        :raises ValidationFailure: unable to validate the Data packet.
        :raises InterestCanceled: the face to NFD is shut down after sending this Interest.
        """
        if not self.face.running:
            raise NetworkError('cannot send packet before connected')
        if 'signer' in kwargs:
            signer = kwargs['signer']
        elif app_param is not None:
            signer = self.keychain.get_signer(kwargs)
        else:
            signer = None
        if 'interest_param' in kwargs:
            interest_param = kwargs['interest_param']
        else:
            if 'nonce' not in kwargs:
                kwargs['nonce'] = gen_nonce()
            interest_param = InterestParam.from_dict(kwargs)
        interest, final_name = make_interest(name, interest_param, app_param, signer=signer, need_final_name=True)
        future = aio.get_event_loop().create_future()
        node = self._int_tree.setdefault(final_name, InterestTreeNode())
        node.append_interest(future, interest_param)
        self.face.send(interest)
        return self._wait_for_data(future, interest_param.lifetime, final_name, node, validator, need_raw_packet)

    async def _wait_for_data(self, future: aio.Future, lifetime: int, name: FormalName,
                             node: InterestTreeNode, validator: Validator, need_raw_packet: bool):
        lifetime = 100 if lifetime is None else lifetime
        try:
            name, meta_info, content, sig, raw_packet = await aio.wait_for(future, timeout=lifetime/1000.0)
        except aio.TimeoutError:
            if node.timeout(future):
                del self._int_tree[name]
            raise InterestTimeout()
        except aio.CancelledError:
            raise InterestCanceled()
        if validator is None:
            validator = self.data_validator
        if await validator(name, sig):
            if need_raw_packet:
                return name, meta_info, content, raw_packet
            else:
                return name, meta_info, content
        else:
            raise ValidationFailure(name, meta_info, content)

    async def main_loop(self, after_start: Awaitable = None):
        """
        The main loop of NDNApp.

        :param after_start: the coroutine to start after connection to NFD is established.
        """
        async def starting_task():
            for name, route, validator, need_raw_packet, need_sig_ptrs in self._autoreg_routes:
                await self.register(name, route, validator, need_raw_packet, need_sig_ptrs)
            if after_start:
                try:
                    await after_start
                except Exception:
                    self.face.shutdown()
                    raise

        try:
            await self.face.open()
        except (FileNotFoundError, ConnectionError, OSError, PermissionError):
            if after_start:
                if isinstance(after_start, Coroutine):
                    after_start.close()
                elif isinstance(after_start, aio.Task) or isinstance(after_start, aio.Future):
                    after_start.cancel()
            raise
        task = aio.ensure_future(starting_task())
        logging.debug('Connected to NFD node, start running...')
        await self.face.run()
        self.face.shutdown()
        self._clean_up()
        await task

    def _clean_up(self):
        for node in self._int_tree.itervalues():
            node.cancel()
        self._prefix_tree.clear()
        self._int_tree.clear()

    def shutdown(self):
        """
        Manually shutdown the face to NFD.
        """
        logging.info('Manually shutdown')
        self.face.shutdown()

    def run_forever(self, after_start: Awaitable = None) -> bool:
        """
        A non-async wrapper of :meth:`main_loop`.

        :param after_start: the coroutine to start after connection to NFD is established.

        :examples:
            .. code-block:: python3

                app = NDNApp()

                if __name__ == '__main__':
                    app.run_forever(after_start=main())
        """
        task = self.main_loop(after_start)
        try:
            aio.get_event_loop().run_until_complete(task)
            ret = True
        except KeyboardInterrupt:
            logging.info('Receiving Ctrl+C, shutdown')
            ret = False
        finally:
            self.face.shutdown()
        logging.debug('Face is down now')
        return ret

    def route(self, name: NonStrictName, validator: Optional[Validator] = None,
              need_raw_packet: bool = False, need_sig_ptrs: bool = False):
        """
        A decorator used to register a permanent route for a specific prefix.

        This function is non-blocking and can be called at any time.
        If it is called before connecting to NFD, NDNApp will remember this route and
        automatically register it every time when a connection is established.
        Failure in registering this route to NFD will be ignored.

        The decorated function should accept 3 arguments: Name, Interest parameters and ApplicationParameters.

        :param name: the Name prefix for this route.
        :type name: :any:`NonStrictName`
        :param validator: the Validator used to validate coming Interests.
            An Interest without ApplicationParameters and SignatureInfo will be considered valid without
            calling validator.
            Interests with malformed ParametersSha256DigestComponent will be dropped before going into the validator.
            Otherwise NDNApp will try to validate the Interest with the validator.
            Interests which fail to be validated will be dropped without raising any exception.
        :type validator: Optional[:any:`Validator`]
        :param need_raw_packet: if True, pass the raw Interest packet to the callback as a keyword argument
            ``raw_packet``.
        :type need_raw_packet: bool
        :param need_sig_ptrs: if True, pass the Signature pointers to the callback as a keyword argument
            ``sig_ptrs``.
        :type need_sig_ptrs: bool

        :examples:
            .. code-block:: python3

                app = NDNApp()

                @app.route('/example/rpc')
                def on_interest(name: FormalName, param: InterestParam, app_param):
                    pass
        """
        name = Name.normalize(name)

        def decorator(func: Route):
            self._autoreg_routes.append((name, func, validator, need_raw_packet, need_sig_ptrs))
            if self.face.running:
                aio.ensure_future(self.register(name, func, validator, need_raw_packet, need_sig_ptrs))
            return func
        return decorator

    async def register(self, name: NonStrictName, func: Optional[Route], validator: Optional[Validator] = None,
                       need_raw_packet: bool = False, need_sig_ptrs: bool = False) -> bool:
        """
        Register a route for a specific prefix dynamically.

        :param name: the Name prefix for this route.
        :type name: :any:`NonStrictName`
        :param func: the onInterest function for the specified route.
            If ``None``, the NDNApp will only send the register command to forwarder,
            without setting any callback function.
        :type func: Optional[Callable[[:any:`FormalName`, :any:`InterestParam`, Optional[:any:`BinaryStr`]], ``None``]]
        :param validator: the Validator used to validate coming Interests.
        :type validator: Optional[:any:`Validator`]
        :return: ``True`` if the registration succeeded.
        :param need_raw_packet: if True, pass the raw Interest packet to the callback as a keyword argument
            ``raw_packet``.
        :type need_raw_packet: bool
        :param need_sig_ptrs: if True, pass the Signature pointers to the callback as a keyword argument
            ``sig_ptrs``.
        :type need_sig_ptrs: bool

        :raises ValueError: the prefix is already registered.
        :raises NetworkError: the face to NFD is down now.
        """
        name = Name.normalize(name)
        if func is not None:
            node = self._prefix_tree.setdefault(name, PrefixTreeNode())
            if node.callback:
                raise ValueError(f'Duplicated registration: {Name.to_str(name)}')
            node.callback = func
            node.extra_param = {'raw_packet': need_raw_packet, 'sig_ptrs': need_sig_ptrs}
            if validator:
                node.validator = validator

        # Fix the issue that NFD only allows one packet signed by a specific key for a timestamp number
        async with self._prefix_register_semaphore:
            try:
                _, _, reply = await self.express_interest(make_command('rib', 'register', name=name), lifetime=1000)
                ret = parse_response(reply)
                if ret['status_code'] != 200:
                    logging.error(f'Registration for {Name.to_str(name)} failed: '
                                  f'{ret["status_code"]} {bytes(ret["status_text"]).decode()}')
                    return False
                else:
                    logging.debug(f'Registration for {Name.to_str(name)} succeeded: '
                                  f'{ret["status_code"]} {bytes(ret["status_text"]).decode()}')
                    return True
            except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure) as e:
                logging.error(f'Registration for {Name.to_str(name)} failed: {e.__class__.__name__}')
                return False

    async def unregister(self, name: NonStrictName) -> bool:
        """
        Unregister a route for a specific prefix.

        :param name: the Name prefix.
        :type name: :any:`NonStrictName`
        """
        name = Name.normalize(name)
        del self._prefix_tree[name]
        try:
            await self.express_interest(make_command('rib', 'unregister', name=name), lifetime=1000)
            return True
        except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure):
            return False

    def set_interest_filter(self, name: NonStrictName, func: Route,
                                  validator: Optional[Validator] = None, need_raw_packet: bool = False,
                                  need_sig_ptrs: bool = False):
        """
        Set the callback function for an Interest prefix without sending a register command to the forwarder.

        .. note::
            All callbacks registered by ``set_interest_filter`` are removed when disconnected from
            the the forwarder, and will not be added back after reconnection.
            This behaviour is the same as ``register``.
            Therefore, it is strongly recommended to use ``route`` for static routes.
        """
        name = Name.normalize(name)
        node = self._prefix_tree.setdefault(name, PrefixTreeNode())
        if node.callback:
            raise ValueError(f'Duplicated registration: {Name.to_str(name)}')
        node.callback = func
        node.extra_param = {'raw_packet': need_raw_packet, 'sig_ptrs': need_sig_ptrs}
        if validator:
            node.validator = validator

    def unset_interest_filter(self, name: NonStrictName):
        """
        Remove the callback function for an Interest prefix without sending an unregister command.

        .. note::
            ``unregister`` will only remove the callback if the callback's name matches exactly
            the route's name.
            This is because there may be one route whose name is the prefix of another.
            To avoid cancelling unexpected routes, neither ``unregister`` nor ``unset_interest_filter``
            behaves in a cascading manner.
            Please remove callbacks manually.
        """
        name = Name.normalize(name)
        del self._prefix_tree[name]

    def _on_nack(self, name: FormalName, nack_reason: int):
        node = self._int_tree[name]
        if node:
            if node.nack_interest(nack_reason):
                del self._int_tree[name]

    async def _on_data(self, name: FormalName, meta_info: MetaInfo,
                       content: Optional[BinaryStr], sig: SignaturePtrs, raw_packet):
        clean_list = []
        for prefix, node in self._int_tree.prefixes(name):
            if node.satisfy((name, meta_info, content, sig, raw_packet), prefix != name):
                clean_list.append(prefix)
        for prefix in clean_list:
            del self._int_tree[prefix]

    async def _on_interest(self, name: FormalName, param: InterestParam,
                           app_param: Optional[BinaryStr], sig: SignaturePtrs, raw_packet: BinaryStr):
        trie_step = self._prefix_tree.longest_prefix(name)
        if not trie_step:
            logging.warning('No route: %s' % name)
            return
        node = trie_step.value
        if node.callback is None:
            logging.warning('No callback: %s' % name)
            return
        if app_param is not None or sig.signature_info is not None:
            if not await params_sha256_checker(name, sig):
                logging.warning('Drop malformed Interest: %s' % name)
                return
        if sig.signature_info is not None:
            validator = node.validator if node.validator else self.int_validator
            valid = await validator(name, sig)
        else:
            valid = True
        if not valid:
            logging.warning('Drop unvalidated Interest: %s' % name)
            return
        if node.extra_param:
            kwargs = {}
            if node.extra_param.get('raw_packet', False):
                kwargs['raw_packet'] = raw_packet
            if node.extra_param.get('sig_ptrs', False):
                kwargs['sig_ptrs'] = sig
            node.callback(name, param, app_param, **kwargs)
        else:
            node.callback(name, param, app_param)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import logging
from ..encoding import FormalName, Name, BinaryStr, InterestParam
from ..name_tree import NameTrie
from .schema_tree import MatchedNode
from . import policy


class MemoryCache:
    """
    MemoryCache is a simple cache class that supports searching and storing Data packets in the memory.
    """
    def __init__(self):
        self.data = NameTrie()

    async def search(self, name: FormalName, param: InterestParam):
        """
        Search for the data packet that satisfying an Interest packet with name specified.

        :param name: the Interest name.
        :param param: the parameters of the Interest. Not used in current implementation.
        :return: a raw Data packet or None.
        """
        try:
            return next(self.data.itervalues(prefix=name, shallow=True))
        except KeyError:
            logging.debug(f'Cache miss: {Name.to_str(name)}')
            return None

    async def save(self, name: FormalName, packet: BinaryStr):
        """
        Save a Data packet with name into the memory storage.

        :param name: the Data name.
        :param packet: the raw Data packet.
        """
        logging.debug(f'Cache save: {Name.to_str(name)}')
        self.data[name] = bytes(packet)


class MemoryCachePolicy(policy.Cache):
    """
    MemoryCachePolicy stores Data packets in memory.
    """
    def __init__(self, cache):
        super().__init__()
        self.cache = cache

    async def search(self, match: MatchedNode, name: FormalName, param: InterestParam):
        return await self.cache.search(name, param)

    async def save(self, match: MatchedNode, name: FormalName, packet: BinaryStr):
        await self.cache.save(name, packet)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import Optional, Callable, Any, Coroutine
from .encoding import FormalName, MetaInfo, BinaryStr, InterestParam, SignaturePtrs


Route = Callable[[FormalName, InterestParam, Optional[BinaryStr]], None]
r"""An OnInterest callback function for a route."""

Validator = Callable[[FormalName, SignaturePtrs], Coroutine[Any, None, bool]]
r"""A validator used to validate an Interest or Data packet."""


class NetworkError(Exception):
    """
    Raised when trying to send a packet before connecting to NFD.
    """
    pass


class InterestTimeout(Exception):
    """
    Raised when an Interest times out.
    """
    pass


class InterestCanceled(Exception):
    """
    Raised when an Interest is cancelled due to the loss of connection to NFD.
    """
    pass


class InterestNack(Exception):
    """
    Raised when receiving a NetworkNack.

    :ivar reason: reason for Nack.
    :vartype reason: int
    """
    reason: int

    def __init__(self, reason: int):
        self.reason = reason


class ValidationFailure(Exception):
    """
    Raised when failing to validate a Data packet.

    :ivar name: the Name of Data.
    :vartype name: :any:`FormalName`
    :ivar meta_info: the MetaInfo.
    :vartype meta_info: :any:`MetaInfo`
    :ivar content: the Content of Data.
    :vartype content: Optional[:any:`BinaryStr`]
    """
    name: FormalName
    meta_info: MetaInfo
    content: Optional[BinaryStr]

    def __init__(self, name: FormalName, meta_info: MetaInfo, content: Optional[BinaryStr]):
        self.name = name
        self.meta_info = meta_info
        self.content = content



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import asyncio as aio
from pygtrie import Trie
from typing import List, Optional, Tuple, Dict
from .encoding import InterestParam, FormalName
from .types import InterestNack, Validator, Route


class NameTrie(Trie):
    def _path_from_key(self, key: FormalName) -> FormalName:
        # bytes(x) will copy x if x is memoryview or bytearray but will not copy bytes
        return [x if isinstance(x, memoryview) and x.readonly else bytes(x)
                for x in key]

    def _key_from_path(self, path: FormalName) -> FormalName:
        return path


class InterestTreeNode:
    pending_list: List[Tuple[aio.Future, int, bool, bool]] = None

    def __init__(self):
        self.pending_list = []

    def append_interest(self, future: aio.Future, param: InterestParam):
        self.pending_list.append((future, param.lifetime, param.can_be_prefix, param.must_be_fresh))

    def nack_interest(self, nack_reason: int) -> bool:
        for future, _, _, _ in self.pending_list:
            future.set_exception(InterestNack(nack_reason))
        return True

    def satisfy(self, data, is_prefix: bool) -> bool:
        exact_match_interest = False
        for future, _, can_be_prefix, _ in self.pending_list:
            if can_be_prefix or not is_prefix:
                future.set_result(data)
            else:
                exact_match_interest = True
        if exact_match_interest:
            self.pending_list = [ele for ele in self.pending_list if not ele[2]]
            return False
        else:
            return True

    def timeout(self, future: aio.Future):
        self.pending_list = [ele for ele in self.pending_list if ele[0] is not future]
        return not self.pending_list

    def cancel(self):
        for future, _, _, _ in self.pending_list:
            future.cancel()


class PrefixTreeNode:
    callback: Route = None
    validator: Optional[Validator] = None
    extra_param: Dict = None



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
from typing import List
from Cryptodome.Hash import SHA256, HMAC
from ...encoding import Signer, SignatureType, KeyLocator, NonStrictName, VarBinaryStr


class HmacSha256Signer(Signer):
    key_name: NonStrictName
    key_bytes: bytes

    def __init__(self, key_name: NonStrictName, key_bytes: bytes):
        self.key_name = key_name
        self.key_bytes = key_bytes

    def write_signature_info(self, signature_info):
        signature_info.signature_type = SignatureType.HMAC_WITH_SHA256
        signature_info.key_locator = KeyLocator()
        signature_info.key_locator.name = self.key_name

    def get_signature_value_size(self):
        return 32

    def write_signature_value(self, wire: VarBinaryStr, contents: List[VarBinaryStr]) -> int:
        h = HMAC.new(self.key_bytes, digestmod=SHA256)
        for blk in contents:
            h.update(blk)
        wire[:] = h.digest()
        return 32



from builtins import str

from ctypes import *
from ctypes import util

from .runtime import send_message, ObjCInstance
from .cocoatypes import *

######################################################################

# CORE FOUNDATION

cf = cdll.LoadLibrary(util.find_library('CoreFoundation'))

kCFStringEncodingUTF8 = 0x08000100

CFAllocatorRef = c_void_p
CFStringEncoding = c_uint32

cf.CFStringCreateWithCString.restype = c_void_p
cf.CFStringCreateWithCString.argtypes = [CFAllocatorRef, c_char_p, CFStringEncoding]

cf.CFRelease.restype = c_void_p
cf.CFRelease.argtypes = [c_void_p]

cf.CFStringGetLength.restype = CFIndex
cf.CFStringGetLength.argtypes = [c_void_p]

cf.CFStringGetMaximumSizeForEncoding.restype = CFIndex
cf.CFStringGetMaximumSizeForEncoding.argtypes = [CFIndex, CFStringEncoding]

cf.CFStringGetCString.restype = c_bool
cf.CFStringGetCString.argtypes = [c_void_p, c_char_p, CFIndex, CFStringEncoding]

cf.CFStringGetTypeID.restype = CFTypeID
cf.CFStringGetTypeID.argtypes = []

cf.CFAttributedStringCreate.restype = c_void_p
cf.CFAttributedStringCreate.argtypes = [CFAllocatorRef, c_void_p, c_void_p]

# Core Foundation type to Python type conversion functions

def CFSTR(string):
    return ObjCInstance(c_void_p(cf.CFStringCreateWithCString(
            None, string.encode('utf8'), kCFStringEncodingUTF8)))

# Other possible names for this method:
# at, ampersat, arobe, apenstaartje (little monkey tail), strudel,
# klammeraffe (spider monkey), little_mouse, arroba, sobachka (doggie)
# malpa (monkey), snabel (trunk), papaki (small duck), afna (monkey),
# kukac (caterpillar).
def get_NSString(string):
    """Autoreleased version of CFSTR"""
    return CFSTR(string).autorelease()

def cfstring_to_string(cfstring):
    length = cf.CFStringGetLength(cfstring)
    size = cf.CFStringGetMaximumSizeForEncoding(length, kCFStringEncodingUTF8)
    buffer = c_buffer(size + 1)
    result = cf.CFStringGetCString(cfstring, buffer, len(buffer), kCFStringEncodingUTF8)
    if result:
        return str(buffer.value, 'utf-8')

cf.CFDataCreate.restype = c_void_p
cf.CFDataCreate.argtypes = [c_void_p, c_void_p, CFIndex]

cf.CFDataGetBytes.restype = None
cf.CFDataGetBytes.argtypes = [c_void_p, CFRange, c_void_p]

cf.CFDataGetLength.restype = CFIndex
cf.CFDataGetLength.argtypes = [c_void_p]

cf.CFDictionaryGetValue.restype = c_void_p
cf.CFDictionaryGetValue.argtypes = [c_void_p, c_void_p]

cf.CFDictionaryAddValue.restype = None
cf.CFDictionaryAddValue.argtypes = [c_void_p, c_void_p, c_void_p]

cf.CFDictionaryCreateMutable.restype = c_void_p
cf.CFDictionaryCreateMutable.argtypes = [CFAllocatorRef, CFIndex, c_void_p, c_void_p]

cf.CFNumberCreate.restype = c_void_p
cf.CFNumberCreate.argtypes = [CFAllocatorRef, CFNumberType, c_void_p]

cf.CFNumberGetType.restype = CFNumberType
cf.CFNumberGetType.argtypes = [c_void_p]

cf.CFNumberGetValue.restype = c_ubyte
cf.CFNumberGetValue.argtypes = [c_void_p, CFNumberType, c_void_p]

cf.CFNumberGetTypeID.restype = CFTypeID
cf.CFNumberGetTypeID.argtypes = []

cf.CFGetTypeID.restype = CFTypeID
cf.CFGetTypeID.argtypes = [c_void_p]

# CFNumber.h
kCFNumberSInt8Type     = 1
kCFNumberSInt16Type    = 2
kCFNumberSInt32Type    = 3
kCFNumberSInt64Type    = 4
kCFNumberFloat32Type   = 5
kCFNumberFloat64Type   = 6
kCFNumberCharType      = 7
kCFNumberShortType     = 8
kCFNumberIntType       = 9
kCFNumberLongType      = 10
kCFNumberLongLongType  = 11
kCFNumberFloatType     = 12
kCFNumberDoubleType    = 13
kCFNumberCFIndexType   = 14
kCFNumberNSIntegerType = 15
kCFNumberCGFloatType   = 16
kCFNumberMaxType       = 16

def cfnumber_to_number(cfnumber):
    """Convert CFNumber to python int or float."""
    numeric_type = cf.CFNumberGetType(cfnumber)
    cfnum_to_ctype = {kCFNumberSInt8Type:c_int8, kCFNumberSInt16Type:c_int16,
                      kCFNumberSInt32Type:c_int32, kCFNumberSInt64Type:c_int64,
                      kCFNumberFloat32Type:c_float, kCFNumberFloat64Type:c_double,
                      kCFNumberCharType:c_byte, kCFNumberShortType:c_short,
                      kCFNumberIntType:c_int, kCFNumberLongType:c_long,
                      kCFNumberLongLongType:c_longlong, kCFNumberFloatType:c_float,
                      kCFNumberDoubleType:c_double, kCFNumberCFIndexType:CFIndex,
                      kCFNumberCGFloatType:CGFloat}

    if numeric_type in cfnum_to_ctype:
        t = cfnum_to_ctype[numeric_type]
        result = t()
        if cf.CFNumberGetValue(cfnumber, numeric_type, byref(result)):
            return result.value
    else:
        raise Exception('cfnumber_to_number: unhandled CFNumber type %d' % numeric_type)

# Dictionary of cftypes matched to the method converting them to python values.
known_cftypes = { cf.CFStringGetTypeID() : cfstring_to_string,
                  cf.CFNumberGetTypeID() : cfnumber_to_number
                  }

def cftype_to_value(cftype):
    """Convert a CFType into an equivalent python type.
    The convertible CFTypes are taken from the known_cftypes
    dictionary, which may be added to if another library implements
    its own conversion methods."""
    if not cftype:
        return None
    typeID = cf.CFGetTypeID(cftype)
    if typeID in known_cftypes:
        convert_function = known_cftypes[typeID]
        return convert_function(cftype)
    else:
        return cftype

cf.CFSetGetCount.restype = CFIndex
cf.CFSetGetCount.argtypes = [c_void_p]

cf.CFSetGetValues.restype = None
# PyPy 1.7 is fine with 2nd arg as POINTER(c_void_p),
# but CPython ctypes 1.1.0 complains, so just use c_void_p.
cf.CFSetGetValues.argtypes = [c_void_p, c_void_p]

def cfset_to_set(cfset):
    """Convert CFSet to python set."""
    count = cf.CFSetGetCount(cfset)
    buffer = (c_void_p * count)()
    cf.CFSetGetValues(cfset, byref(buffer))
    return set([ cftype_to_value(c_void_p(buffer[i])) for i in range(count) ])

cf.CFArrayGetCount.restype = CFIndex
cf.CFArrayGetCount.argtypes = [c_void_p]

cf.CFArrayGetValueAtIndex.restype = c_void_p
cf.CFArrayGetValueAtIndex.argtypes = [c_void_p, CFIndex]

def cfarray_to_list(cfarray):
    """Convert CFArray to python list."""
    count = cf.CFArrayGetCount(cfarray)
    return [ cftype_to_value(c_void_p(cf.CFArrayGetValueAtIndex(cfarray, i)))
             for i in range(count) ]


kCFRunLoopDefaultMode = c_void_p.in_dll(cf, 'kCFRunLoopDefaultMode')

cf.CFRunLoopGetCurrent.restype = c_void_p
cf.CFRunLoopGetCurrent.argtypes = []

cf.CFRunLoopGetMain.restype = c_void_p
cf.CFRunLoopGetMain.argtypes = []

######################################################################

# APPLICATION KIT

# Even though we don't use this directly, it must be loaded so that
# we can find the NSApplication, NSWindow, and NSView classes.
appkit = cdll.LoadLibrary(util.find_library('AppKit'))

NSDefaultRunLoopMode = c_void_p.in_dll(appkit, 'NSDefaultRunLoopMode')
# NSEventTrackingRunLoopMode = c_void_p.in_dll(appkit, 'NSEventTrackingRunLoopMode')
# NSApplicationDidHideNotification = c_void_p.in_dll(appkit, 'NSApplicationDidHideNotification')
# NSApplicationDidUnhideNotification = c_void_p.in_dll(appkit, 'NSApplicationDidUnhideNotification')

# /System/Library/Frameworks/AppKit.framework/Headers/NSEvent.h
NSAnyEventMask = 0xFFFFFFFF     # NSUIntegerMax

NSKeyDown            = 10
NSKeyUp              = 11
NSFlagsChanged       = 12
NSApplicationDefined = 15

NSAlphaShiftKeyMask         = 1 << 16
NSShiftKeyMask              = 1 << 17
NSControlKeyMask            = 1 << 18
NSAlternateKeyMask          = 1 << 19
NSCommandKeyMask            = 1 << 20
NSNumericPadKeyMask         = 1 << 21
NSHelpKeyMask               = 1 << 22
NSFunctionKeyMask           = 1 << 23

NSInsertFunctionKey   = 0xF727
NSDeleteFunctionKey   = 0xF728
NSHomeFunctionKey     = 0xF729
NSBeginFunctionKey    = 0xF72A
NSEndFunctionKey      = 0xF72B
NSPageUpFunctionKey   = 0xF72C
NSPageDownFunctionKey = 0xF72D

# /System/Library/Frameworks/AppKit.framework/Headers/NSWindow.h
NSBorderlessWindowMask		= 0
NSTitledWindowMask		= 1 << 0
NSClosableWindowMask		= 1 << 1
NSMiniaturizableWindowMask	= 1 << 2
NSResizableWindowMask		= 1 << 3

# /System/Library/Frameworks/AppKit.framework/Headers/NSPanel.h
NSUtilityWindowMask		= 1 << 4

# /System/Library/Frameworks/AppKit.framework/Headers/NSGraphics.h
NSBackingStoreRetained	        = 0
NSBackingStoreNonretained	= 1
NSBackingStoreBuffered	        = 2

# /System/Library/Frameworks/AppKit.framework/Headers/NSTrackingArea.h
NSTrackingMouseEnteredAndExited  = 0x01
NSTrackingMouseMoved             = 0x02
NSTrackingCursorUpdate 		 = 0x04
NSTrackingActiveInActiveApp 	 = 0x40

# /System/Library/Frameworks/AppKit.framework/Headers/NSOpenGL.h
NSOpenGLPFAAllRenderers       =   1   # choose from all available renderers
NSOpenGLPFADoubleBuffer       =   5   # choose a double buffered pixel format
NSOpenGLPFAStereo             =   6   # stereo buffering supported
NSOpenGLPFAAuxBuffers         =   7   # number of aux buffers
NSOpenGLPFAColorSize          =   8   # number of color buffer bits
NSOpenGLPFAAlphaSize          =  11   # number of alpha component bits
NSOpenGLPFADepthSize          =  12   # number of depth buffer bits
NSOpenGLPFAStencilSize        =  13   # number of stencil buffer bits
NSOpenGLPFAAccumSize          =  14   # number of accum buffer bits
NSOpenGLPFAMinimumPolicy      =  51   # never choose smaller buffers than requested
NSOpenGLPFAMaximumPolicy      =  52   # choose largest buffers of type requested
NSOpenGLPFAOffScreen          =  53   # choose an off-screen capable renderer
NSOpenGLPFAFullScreen         =  54   # choose a full-screen capable renderer
NSOpenGLPFASampleBuffers      =  55   # number of multi sample buffers
NSOpenGLPFASamples            =  56   # number of samples per multi sample buffer
NSOpenGLPFAAuxDepthStencil    =  57   # each aux buffer has its own depth stencil
NSOpenGLPFAColorFloat         =  58   # color buffers store floating point pixels
NSOpenGLPFAMultisample        =  59   # choose multisampling
NSOpenGLPFASupersample        =  60   # choose supersampling
NSOpenGLPFASampleAlpha        =  61   # request alpha filtering
NSOpenGLPFARendererID         =  70   # request renderer by ID
NSOpenGLPFASingleRenderer     =  71   # choose a single renderer for all screens
NSOpenGLPFANoRecovery         =  72   # disable all failure recovery systems
NSOpenGLPFAAccelerated        =  73   # choose a hardware accelerated renderer
NSOpenGLPFAClosestPolicy      =  74   # choose the closest color buffer to request
NSOpenGLPFARobust             =  75   # renderer does not need failure recovery
NSOpenGLPFABackingStore       =  76   # back buffer contents are valid after swap
NSOpenGLPFAMPSafe             =  78   # renderer is multi-processor safe
NSOpenGLPFAWindow             =  80   # can be used to render to an onscreen window
NSOpenGLPFAMultiScreen        =  81   # single window can span multiple screens
NSOpenGLPFACompliant          =  83   # renderer is opengl compliant
NSOpenGLPFAScreenMask         =  84   # bit mask of supported physical screens
NSOpenGLPFAPixelBuffer        =  90   # can be used to render to a pbuffer
NSOpenGLPFARemotePixelBuffer  =  91   # can be used to render offline to a pbuffer
NSOpenGLPFAAllowOfflineRenderers = 96 # allow use of offline renderers
NSOpenGLPFAAcceleratedCompute =  97   # choose a hardware accelerated compute device
NSOpenGLPFAOpenGLProfile      =  99   # specify an OpenGL Profile to use
NSOpenGLPFAVirtualScreenCount = 128   # number of virtual screens in this format

NSOpenGLProfileVersionLegacy  = 0x1000    # choose a Legacy/Pre-OpenGL 3.0 Implementation
NSOpenGLProfileVersion3_2Core = 0x3200    # choose an OpenGL 3.2 Core Implementation
NSOpenGLProfileVersion4_1Core = 0x4100    # choose an OpenGL 4.1 Core Implementation

NSOpenGLCPSwapInterval        = 222


# /System/Library/Frameworks/ApplicationServices.framework/Frameworks/...
#     CoreGraphics.framework/Headers/CGImage.h
kCGImageAlphaNone                   = 0
kCGImageAlphaPremultipliedLast      = 1
kCGImageAlphaPremultipliedFirst     = 2
kCGImageAlphaLast                   = 3
kCGImageAlphaFirst                  = 4
kCGImageAlphaNoneSkipLast           = 5
kCGImageAlphaNoneSkipFirst          = 6
kCGImageAlphaOnly                   = 7

kCGImageAlphaPremultipliedLast = 1

kCGBitmapAlphaInfoMask              = 0x1F
kCGBitmapFloatComponents            = 1 << 8

kCGBitmapByteOrderMask              = 0x7000
kCGBitmapByteOrderDefault           = 0 << 12
kCGBitmapByteOrder16Little          = 1 << 12
kCGBitmapByteOrder32Little          = 2 << 12
kCGBitmapByteOrder16Big             = 3 << 12
kCGBitmapByteOrder32Big             = 4 << 12

# NSApplication.h
NSApplicationPresentationDefault = 0
NSApplicationPresentationHideDock = 1 << 1
NSApplicationPresentationHideMenuBar = 1 << 3
NSApplicationPresentationDisableProcessSwitching = 1 << 5
NSApplicationPresentationDisableHideApplication = 1 << 8

# NSRunningApplication.h
NSApplicationActivationPolicyRegular = 0
NSApplicationActivationPolicyAccessory = 1
NSApplicationActivationPolicyProhibited = 2



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 The python-ndn authors
#
# This file is part of python-ndn.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------
import time
from random import randint


def timestamp():
    """
    Generate a timestamp number.

    :return: the time in milliseconds since the epoch as an integer
    """
    return int(time.time() * 1000)


def gen_nonce():
    """
    Generate a random nonce.

    :return: a random 32-bit unsigned integer.
    """
    return randint(1, 2 ** 32 - 1)


def gen_nonce_64():
    """
    Generate a random 64-bit nonce.

    :return: a random 64-bit unsigned integer.
    """
    return randint(1, 2 ** 64 - 1)



from .sha256_rsa_signer import Sha256WithRsaSigner
from .sha256_digest_signer import DigestSha256Signer
from .sha256_ecdsa_signer import Sha256WithEcdsaSigner
from .sha256_hmac_signer import HmacSha256Signer


__all__ = ['Sha256WithRsaSigner', 'DigestSha256Signer', 'Sha256WithEcdsaSigner', 'HmacSha256Signer']
